# usage: perl generate.pl <number of clients>
#
# creates a set of files for SPECweb99 benchmark and a meta-file which contains pairs <fileName, fileProb>
# where "fileName" is the name of the file and "fileProb" is the probability of access to that file

# Owner server is the initial owner of the web resources in Web++ tests
#  This may be set to a different machine from NAME_SERVER, which is
#  the coordinator that sends phase change messages to others and
#  provides the list of active servers to the clients.
#  e.g. $owner_server = 'lilac.ece.northwestern.edu:8080';
#  e.g. $owner_server = 'london.cs.uni-sb.de:8080';
$owner_server = 'cs.sookmyung.ac.kr:8090';

$clients = $ARGV[0];
$dirs = int(25 + (((400000.0/122000.0) * $clients) / 5.0));
# if the total size is too large, set $dirs manually to whatever value desired
# $dirs = 2; -- this is an example

$dirs = 1;
# If generate_files is 1, then test files are generated, in addition
#  to probability generation. Otherwise, only probabilities are generated
$generate_files = 0;

$expectedSize = 5083136 * $dirs;
$noofphases = 3;
print( "expected total size = $expectedSize \n" );

sub genDirProb;
sub genFile;
sub sum;
sub genMetaFile;

genDirProb($dirs);
for( $dir = 0; $dir < $dirs; $dir++ ){
    print( "generating dir $dir \n" );

    for( $class = 0; $class < 4; $class++ ){
	for( $fileId = 0; $fileId < 9; $fileId++ ){
	    genFile( $dir, $class, $fileId );
        }
    }
}
genMetaFile();



#######################################################################################
#
sub genFile
{
    my( $dir, $class, $fileId ) = @_;
    my( $fileName, @incr, @base, $size, @classProb, @fileIdProb, $fileProb, $i, $buf );
   
    for ($phase=0; $phase<$noofphases; $phase++) {

    $fileName = "file_" . "$phase" . "d$dir" . "c$class" . "id$fileId.txt";
    if (generate_files > 0) {
      open( FILE, ">$fileName" ) || die( "could not open $fileName \n" );
    }
  
    @incr = (100, 1000, 10000, 100000);
    @base = (0, 1000, 10000, 100000);
    $size = $base + ($fileId + 1) * $incr[$class]; # in B
    for( $i = 0; $i < 50; $i++ ){
	$buf = $buf . "#\n";
    }
    if (generate_files > 0) {
      for( $i = 0; $i < $size; $i += 100 ){
	print( FILE $buf );
      }
      close( FILE );
    }
    $file2{$fileName} = $fileProb;
    }

    $fileName = "file_" . "d$dir" . "c$class" . "id$fileId.txt";
    @classProb = (0.35, 0.5, 0.14, 0.01);
    @fileIdProb = (0.039, 0.059, 0.088, 0.177, 0.353, 0.118, 0.071, 0.050, 0.044);   
    $fileProb = $dirProb[$dir] * $classProb[$class] * $fileIdProb[$fileId];

    $file{$fileName} = $fileProb;
}


#######################################################################################
#
sub sum
{
    my @l = @_;
    my $sum = 0;
    my $p;

    foreach $p (@l) {
	$sum += $p;
    }

    return $sum;
}


#######################################################################################
#
sub genDirProb
{
    my ($dirs) = @_;
    my( $i, $H );

    for( $i = 0; $i < $dirs; $i++ ){
	$dirProb[$i] = 1. / ($i + 1);
    }

    $H = sum( @dirProb );

    for( $i = 0; $i < $dirs; $i++ ){
	$dirProb[$i] /= $H;
    }
}


#######################################################################################
#
sub genMetaFile
{
    my $f;

#   Read all machine names
    open(INFILE1, "<..//webpp/replication/servers.ser");
    $count = 0;
    while (<INFILE1>) {
	my($line) = $_;
	chomp($line);	# strip the end-of-line character
	$line =~ s/ /:/g;	# replace space with ':'
	$line =~ s/#//g;	# remove any preceeding '#'s
	if ($count > 0) { $fnames1 .= " "; $fnames2 .= " "; $fnames3 .= " "; }
	$fnames1 .= "http://" . $line . "/servlet/webppserv.Webpp/index.html";
	$fnames2 .= "http://" . $line . "/servlet/webppserv.Webpp/indexWebpp.html";
	$fnames3 .= "http://" . $line . "/servlet/webppserv.Webpp/dummy.html";
	$count++;
    }
    close(INFILE1);
    open(INFILE2, "<..//webpp/replication/clients.ser");
    while (<INFILE2>) {
        my($line) = $_;
        chomp($line);   # strip the end-of-line character
	$line =~ s/#//g;	# remove any preceeding '#'s
	$line =~ s/ /:/g;	# replace space with ':'
	$fnames1 .= " http://" . $line . "/servlet/webppserv.Webpp/index.html";
        $fnames2 .= " http://" . $line . "/servlet/webppserv.Webpp/indexWebpp.html";
        $fnames3 .= " http://" . $line . "/servlet/webppserv.Webpp/dummy.html";
    }
    close(INFILE2);

    open( FILE, ">testProbabilities.txt" );
    open( FILE2, ">../webpp/replication/replicationDir.ser.start" );
    #print ( FILE2 "index.html NULL http://$owner_server/servlet/webppserv.Webpp/index.html\n");
    #print ( FILE2 "indexWebpp.html NULL http://$owner_server/servlet/webppserv.Webpp/indexWebpp.html\n");
    print ( FILE2 "index.html NULL $fnames1\n");
    print ( FILE2 "indexWebpp.html NULL $fnames2\n");
    print ( FILE2 "dummy.html NULL $fnames3\n");
    foreach $f (keys(%file)) {
	#print( FILE "$f \t $file{$f} \n" );
	printf ( FILE "%s %12.10f\n", $f, $file{$f});
    }
    foreach $f (keys(%file2)) {
        print ( FILE2 "$f NULL http://$owner_server/servlet/webppserv.Webpp/$f\n" );
    }
}

