package webppclient;

// Web++ Client Cache
// Cache class: keeps track of client cache documents

import java.util.*;
import java.lang.*;
import java.io.*;

final class Cache extends Hashtable implements Serializable
{

  Cache()
  { super(); }

  // ReplaceSeparator replaces the file separator characters
  // in the given filename to the current system's file separator.
  synchronized String ReplaceSeparator(String fname, char sepold, char sepnew)
  {
    fname.replace(sepold, sepnew);
    return fname;
  }

  // inCache returns true or false depending on whether a given
  // file is cached or not.
  synchronized boolean inCache(String fname)
  {

    // The following if-else stmts check the cache index contents.
    // If the Webpp.useCacheIndex is set to true, then the decision
    // is made using the cache index. Otherwise, cache index does not
    // have any effect. The rest of this function uses the \Temp directory
    // which is used by the applet to store the downloaded files.
    if (containsKey(fname))
    {
      //System.out.println("In CacheIndex: " + fname);
      if (Webpp.useCacheIndex) return true;
    }
    else
    {
      //System.out.println("Not in CacheIndex: " + fname);
      if (Webpp.useCacheIndex) {
	return false;
      }
    }

    // If the function execution comes up to this point, it means cache
    // index is not being used. So, now we are going to check if the
    // given file is in the \Temp directory. If, it is there, it means
    // that the file is in cache. Otherwise, it is not in cache.

    // Get rid of any directory names in the filename.
    int i = fname.lastIndexOf('/');
    if (i != -1) fname = fname.substring(i+1);
    //System.out.println("Mehmet: new fname=" + fname);

    // Check if the file exists in Temp directory.
    File fd = new File(Webpp.baseDir + fname);
    if (fd.exists()) {
      //System.out.println("In Cache directory: " + fname);
      return true;
    }
    else {
      //System.out.println("Not in Cache directory: " + fname);
      return false;
    }
  }

  // putCache inserts the file to the cache index with the current time
  // as the last update time. If the file was already in the cache,
  // then only the last update time will be modified.
  synchronized void putCache(String fname)
  {
    long ftime = System.currentTimeMillis();
    //System.out.println("Putting cache: " + fname);
    CacheItem item = new CacheItem(fname, ftime);
    put(fname, item);
  }

  // getCache returns the last update time of the cached file
  synchronized long getCache(String fname)
  {
   CacheItem item = (CacheItem) get(fname);
   return item.ftime;
  }

  // checkEmbResources receives a vector containing the list of
  // embedded resources related to the current URL being requested.
  // It determines which ones of those resources are in client
  // cache and returns a string containing those resources' names.
  synchronized String checkEmbResources(Vector embRes)
  {
    String fname;
    String embStr = "?;";
    if (embRes == null) {
      System.out.println("Error: embRes is null in Cache.checkEmbResources");
      return null;
    }
    if (embRes.isEmpty()) {
      System.out.println("Error: embRes is empty in Cache.checkEmbResources");
      return null;
    }
    //System.out.println("embRes: " + embRes.toString());
    //if (Webpp.ClientHasParentDoc)
    //  embStr = embStr + "parent;";
    Enumeration e = embRes.elements();
    int count = 0, hasCount = 0;
    for (; e.hasMoreElements(); ) {
      count++;
      fname = (String) e.nextElement();
      //System.out.println("Mehmet: cache check for " + fname);
      if (inCache(fname)) {
	//System.out.println("Mehmet: in cache");
        embStr = embStr + fname + ";";
	hasCount++;
      }
      //else System.out.println("Mehmet: not in cache");
    }
    if (embStr.length() == 2) return null;	// client has none of the emb resources

    // Mehmet: the following is just a short-cut to make processing faster on servlet
    //	when the client's cache has all the embedded resources. I commented it out
    //	later on because it is not safe to assume that the servlet does not have to
    //	send anything when the client's cache has all the embedded resources and the parent
    //	document. Because, the list of embedded resources in the HREF of the current document
    //	shown on the browser might be stale. Even if the client has the most recent copy
    //	of the document pointed by the HREF, the current document which has an HREF link to
    //	the next document could be stale.
    //if (hasCount == count) embStr = "?ALL";	// client has all emb resources

    return embStr;				// client has some of the emb resources
  }
}

// CacheItem class: used in client cache for the 'value' objects
// of <key,value> pairs in the cache.

final class CacheItem implements Serializable
{
  String fname;
  long ftime;

  public CacheItem(String fn, long ft)
  {
    fname = fn;
    ftime = ft;
  }

  private CacheItem()
  { ; }
}
