package webppclient;
// Web++ Client TestRunner thread class

import java.io.*;
import java.lang.*;
import java.util.*;

class TestRunner extends Thread implements Runnable
{
  private Webpp applet_;
  static final int numTestPhases = 3;
  String resourceName;	// name of resource to be requested at each iteration
  String resourceLocations;	// locations of a resource
  String selectedLocation;	// selected (best) location to request the resource
  int count;	// counts the request number

  public TestRunner(Webpp app)
  {
	applet_ = app;	// reference to parent applet
  }

  public String extractServersFromLocations(String resourceLocations)
  {
	int pos, pos2;
	String result = null, tmp, location;
	pos = resourceLocations.indexOf(' ');
	if (pos < 0) return null;
	else tmp = resourceLocations.substring(pos+1);	// eliminate NULL at the beginning of replication directory record that is received from host server
	StringTokenizer st = new StringTokenizer(tmp);
	for (int i=0; st.hasMoreElements(); i++) {
          location = ((String) st.nextElement());
	  pos = location.indexOf("//");
	  if (pos < 0) continue;
	  pos2 = location.indexOf('/', pos+2);
	  if (pos2 < 0) continue;
	  if (result == null) result = location.substring(pos+2, pos2);
	  else result += " " + location.substring(pos+2, pos2);
        }
	return result;
  }

  public void run()
  {
        String serverName, port;
        Enumeration enum;
        ThreadGroup msgTG    = new ThreadGroup("Message Group");
	count = 0;
	long t0, t1;	// used for measuring time
	String result = null;	// used to keep the result of getUrl call

        yield();        // give the caller a chance to go on

	if ((System.currentTimeMillis() - applet_.testStartTime) < 6000000) {
          System.out.println("Sleeping before starting the tests.");
	  // Wait 30 seconds before starting the test
          try {
            sleep(30000);
          } catch (InterruptedException ie) {}

	  // Sleep a randomly chosen amount of time
	  double sleepdecision = applet_.myrandom.nextDouble();
	  long sleeptime = 0;
	  if (sleepdecision > 0.9) sleeptime = 90000;
	  else if (sleepdecision > 0.8) sleeptime = 80000;
	  else if (sleepdecision > 0.7) sleeptime = 70000;
	  else if (sleepdecision > 0.6) sleeptime = 60000;
	  else if (sleepdecision > 0.5) sleeptime = 50000;
	  else if (sleepdecision > 0.4) sleeptime = 40000;
	  else if (sleepdecision > 0.3) sleeptime = 30000;
	  else if (sleepdecision > 0.2) sleeptime = 20000;
	  else if (sleepdecision > 0.1) sleeptime = 10000;
	  if (sleeptime > 0) {
	    System.out.println("Sleeping " + sleeptime + " msec more.");
	    try {
	      sleep(sleeptime);
	    } catch (InterruptedException ie) {}
	  }
	}

	while (applet_.testPhase < numTestPhases) {
	  count++;
	  // Re-select home server after every 50 requests
          if (count % 50 == 0) {
	    System.out.println("Refreshing HostServer selection...");
	    applet_.chooseHostServer();
	  }	// if count...
	  applet_.logStr = Long.toString(System.currentTimeMillis() - applet_.testStartTime);
	  applet_.logStr += " Client " + applet_.server + " ClientRequest ";
	  resourceName = pickResource();
	  applet_.logStr += resourceName;
	  t0 = System.currentTimeMillis();
	  resourceLocations = getResourceLocations(resourceName);
	  t1 = System.currentTimeMillis() - t0;
	  if (resourceLocations != null) {
	    String tmpRes = extractServersFromLocations(resourceLocations);
	    if (tmpRes == null) tmpRes = "null";
	    applet_.writeToLog(Long.toString(System.currentTimeMillis() - applet_.testStartTime) + " Client " + applet_.server + " LocationsRequest " + resourceName + " " + Long.toString(t1) + " " + tmpRes + "\n");
	  }
	  if (resourceLocations == null || applet_.hostServer == null) {
	    System.out.println("Have to re-select a HostServer...");
	    // Re-select home server if there was a problem in reaching the current one
	    applet_.writeToLogWithTime("NotResponding " + applet_.hostServer + " LocationsRequest");
	    applet_.chooseHostServer();
	  }	// if resourceLocations...
	  // Now, send a request for a randomly selected resource
	  // e.g. NULL http://begonia.ece.northwestern.edu:8080/servlet/webppserv.Webpp/index.html

	  if (resourceLocations != null && resourceLocations.length() > 0) {
	    try {
	      result = applet_.getUrl(resourceLocations);
	    } catch ( Exception e ) {
	      result = null;
	      System.out.println( "Error: Exception in TestRunner:run " + e.getMessage() );
	      //Runtime.getRuntime().exit( -1 );
	    } 
	    if (result != null) {
	      applet_.logStr += " " + Integer.toString(count);      // add request count to log record
	      applet_.writeToLog(applet_.logStr + "\n");
	    }	// if (result != null) ...
	  }	// if resourceLocations...
	  yield();	// give a chance to parent thread to check for incoming requests, if any

	  // Wait for a while before sending the next request
	  if (applet_.testPhase < numTestPhases) {
	    long waittime = 5000;	// 5 seconds if warm up period is over
	    // Long inter-request wait time is used only during warm up period
	    // If the client is restarted during the test, current test clock
	    // will most probably be larger than 6 million msecs.
	    if ((System.currentTimeMillis() - applet_.testStartTime) < 6000000)
	    {
	      if (count < 10) waittime = 20000;
	      else if (count < 20) waittime = 10000;
	    }
	    //System.out.println("Sleeping " + waittime + " Zzzzzzzz.");
            try {
              sleep(waittime);
            } catch (InterruptedException ie) {}
	  }
	}	// while testPhase...

	System.out.println("Test is completed at " + applet_.server);
	applet_.writeToLogWithTime("End");
	try {
	  applet_.logfile.close();
        } catch (IOException e) {
          System.out.println("Error: cannot close log file");
        }

	applet_.destroy();	// unload client servlet from web server

  }	// run()

  // Pick a resource name to be requested from a Web++ server
  String pickResource() {
	String resourceName;
	double randnum;
	int i;
	randnum = applet_.myrandom.nextDouble();
	for (i=0; randnum > applet_.probabilities[i] && i < applet_.testfcount; i++);
	resourceName = applet_.testfnames[i];
	int pos = resourceName.indexOf('_');
	if (pos < 0) {
	  System.out.println("Error: no _ in filename: " + resourceName);
	}
	else {
	  String tmp = resourceName.substring(0, pos+1) + 
			Integer.toString(applet_.testPhase) +
			resourceName.substring(pos+1);
	  resourceName = tmp;
	}
	System.out.println("\nResource: " + resourceName);
	System.out.println("My name: " + applet_.server + " Count: " + count);
	return resourceName;
  }	// pickResource

  // Contact this Web++ client's HostServer and retrieve the locations
  // of the given resource
  String getResourceLocations(String resourceName) {
	String resourceLocations = null;
	String url = null;
	if (applet_.hostServer != null)
	  url = new String("http://" + applet_.hostServer);
	else
	  url = new String("http://" + applet_.NAME_SERVER);

	//System.out.println("URL for Locations request: " + url);
	try {
	  resourceLocations = applet_.doPostRequest(url, "Locations&" + resourceName);
	} catch (Exception e) {
	  resourceLocations = null;
	  System.out.println("Error: Cannot send Locations request to NAME_SERVER: "+ e);
	}
	if (resourceLocations == null && applet_.hostServer == null) {
	  System.out.println("Re-submitting with IP address");
	  url = new String("http://" + applet_.NAME_SERVER_IP);
	  try {
            resourceLocations = applet_.doPostRequest(url, "Locations&" + resourceName);
          } catch (Exception e) {
            resourceLocations = null;
            System.out.println("Error: Cannot send Locations request to NAME_SERVER: "+ e);
	    applet_.writeToLogWithTime("NotResponding " + applet_.NAME_SERVER + " LocationsRequest");
            return null;
          }
	}
	//System.out.println("Locations received: " + resourceLocations);
	return resourceLocations;
  }

}
