package webppclient;
// Web++ Client

// The initialization is done and tests are started when
// one of the following happens:
//  - A get request is received and the initialization has not been
//	done before
//  - A post request is received and it sets the TestPhase to zero
//	(note that in this case, the initialization is done and
//	the tests are started even if those were already done before.
//	Thus, a TestPhase call with phase number 0 is used for
//	restarting the tests.

import java.net.*;
import java.io.*;
import java.util.*;
import java.applet.*;
import java.awt.*;

import javax.servlet.*;
import javax.servlet.http.*;


///////////////////////////////////////////////////////////////////////////////
// 
//
///////////////////////////////////////////////////////////////////////////////
public class Webpp extends HttpServlet
{

    static final int BUFFER_SIZE = 8192;   // buffer size used to read files
    static final boolean useCache = false;	// Mehmet: enable if want to use cache
    static long testStartTime = 0;
    static int testPhase = 0;
    static Random myrandom;
    static String server = null;	// name of this machine
    static String logStr;	// string to be written to log
    static String logfname = "public_html/logfile.txt";	// filename for the log file
    //static PrintWriter logfile = null;
    static RandomAccessFile logfile = null;

    // package data members ===================================================
    static LatTable lat;
    static Cache cache;				// Mehmet
    static RespTimeTable respTimes;		// Mehmet
    static String baseDir;
    static String latFilename;
    static String cacheFileName;		// Mehmet
    static String respTimeFileName;		// Mehmet
    static boolean useCacheIndex = true;	// Mehmet
    static boolean ClientHasParentDoc;	// Mehmet: set to true if parent HTML doc is in client cache
    static int os;
    //static int browser;
    static boolean trace = false;
    static boolean timing = true;
    static boolean dummyRequestsStarted = false;	// Mehmet: used when determining the closest (host) server
    static boolean dummyRequestsFinished = false;	// Mehmet: used when determining the closest (host) server
    static String hostServer = null;			// Mehmet: chosen host server URL including its port number
    static String serverNames = null;		// Mehmet: names and port numbers of all servers
    static String dummyFileName = null;		// Mehmet: used when returned dynamic results to user (e.g. when choosing host server)
    static long responseTime = 0;	// Mehmet: stores the response time of current request, if successful
    static String prevResponseTime = null;	// Mehmet: stores the URL and response time of previous request
    static double[] probabilities = null;	// Mehmet: request probabilities of test files
    static String[] testfnames = null;		// Mehmet: filenames of the test files, whose probabilities are kept in probabilities array
    static int testfcount = 0;			// Mehmet: # of test files
    
    // package constants ======================================================
    final static int TTL           = 900000; // TTL for stat entries (msec)
    final static long FAIL_PERIOD_INI = 180000; // initial fail period (msec)
    final static int STAT_SIZE     =  5;  // statistics array size
    final static double PERCENTILE = 55;  // percentile for server selection
    final static boolean ASSERT_ON = true;
    final static int UNIX          = 21;
    final static int WIN           = 22;
    final static int EXPLORER      = 23;
    final static int NAVIGATOR     = 24;
    final static int BUF_LEN       = 8192; // size of a communications buffer
    final static String NAME_SERVER = "begonia.ece.northwestern.edu:8080";	// Name server (coordinator)
    final static String NAME_SERVER_IP = "129.105.5.141:8080";	// IP address of the name server
    
    // debug flags ============================================================
    // prints the selected server, dumps latency table and informs about asynch
    // samples
    final static boolean SERVER_CHOICE_DEBUG = false;
    
    // contains body of main for debugging
    final static boolean MAIN_DEBUG = false;

    public void init (ServletConfig config) throws ServletException
    {
	super.init(config);
	testPhase = 0;
    }	// init fcn.

    public void initWebppClient() throws ServletException
    {
	//System.out.println("Mehmet: in Webpp init");
	testStartTime = getTestTimeFromNameServer();
        if (testStartTime == 0) testStartTime = System.currentTimeMillis();
	testPhase = 0;
	//myrandom = new Random(1);
	myrandom = new Random(testStartTime);
	oldinit();	// call the old init routines inherited from applet version

	// Read the request probabilities for the test files
	String line, fname, tmp;
	double prob = 0;
	int pos;
	BufferedReader reader = null;
	try {
	  reader = new BufferedReader(new FileReader("public_html/testProbabilities.txt"));
	} catch (FileNotFoundException e) {
	  System.out.println("Error: FileNotFoundException in init while reading testProbabilities.txt");
	  return;
	}
	probabilities = new double[1000];
	testfnames = new String[1000];
	int count = 0;
	// Read the first line
	try {
	  line = reader.readLine();
	} catch (IOException e) {
	  System.out.println("Error: IOException while reading testProbabilities.txt");
	  return;
	}
	while (line != null) {
	  pos = line.indexOf(" ");
	  if (pos < 0 ) {
	    System.out.println("Error: cannot parse the input line from testProbabilities.txt");
	    return;
	  }
	  fname = line.substring(0, pos);
	  tmp = line.substring(pos+1);
	  prob += (new Double(tmp)).doubleValue();
	  probabilities[count] = prob;
	  testfnames[count] = fname;
	  count++;
	  //System.out.println("Count: " + count + "   prob: " + prob);
	  // Read the next line
	  try {
            line = reader.readLine();
          } catch (IOException e) {
            System.out.println("Error: IOException while reading meta.txt");
            return;
          }
	}
	testfcount = count;
	if (count > 0)
	  probabilities[count-1] = 1.0;	// last probability is rounded to 1

	// Create or open the log file
	File fd = new File( logfname );
        if( fd.exists() ) {
          // Open the existing log file for append
          try {
            logfile = new RandomAccessFile(fd, "rw");
          } catch (IOException e) {
            System.out.println("Error: cannot open logfile for read-write");
          }
          try {
            logfile.seek(logfile.length());     // go to end of file (Append mode)
          } catch (IOException e) {
            System.out.println("Error: cannot seek end of log file");
          }
        }
        else {
          // Create a new log file
          try {
            //logfile = new PrintWriter(new BufferedWriter(new FileWriter(logfname)));
            logfile = new RandomAccessFile(logfname, "rw");
          } catch( IOException e ){
	    System.out.println( "Error: cannot create log file: " + e.getMessage() );
          }
        }

	writeToLogWithTime("Start");

        // If Host server is not chosen yet, choose it
        //if (hostServer == null) chooseHostServer();

        // Mehmet: instead of the above if stmt, I always let the
        //  client choose a host server now (for the automatic tests)
        chooseHostServer();

        // Start TestRunner thread
        Runnable r = new TestRunner(this);
        Thread t = new Thread(r);
        t.start();

    }	// initWebppClient fcn.

    // applet constructor =====================================================
    public void oldinit()
    {
        String osName;
        String par;

        try {
            //par = getParameter( "trace" );
            //if( par != null && par.indexOf( "true" ) != -1 ){
            //    trace = true;
            //} else {
            //    trace = false;
            //}

	    trace = true;
            
            //par = getParameter( "timing" );
            //if( par != null && par.indexOf( "true" ) != -1 ){
            //    timing = true;
            //} else {
            //    timing = false;
            //}

	    timing = true;
            
            osName = System.getProperty( "os.name" );
            if( osName.indexOf( "indow" ) != -1 ){
                os = WIN;
                baseDir = new String( "C:\\TEMP\\" ); 
                if( trace ){
                    System.out.println( "OS: Windows" );
                }
            } else {
                os = UNIX;
                baseDir = new String( "/tmp/" );
                if( trace ){
                    System.out.println( "OS: Unix" );
                }
            }
            
            //par = getParameter( "env" );
            //Assert.assert( par != null );
            //if( par.indexOf( "icrosoft" ) != -1 ){
            //    browser = EXPLORER;
            //    if( trace ){
            //        System.out.println( "Browser: Explorer" );
            //    }
            //} else {
            //    browser = NAVIGATOR;
            //    if( trace ){
            //        System.out.println( "Browser: Navigator" );
            //    }
            //}
           
            latFilename = new String( baseDir + "LatTable.ser" );
            File latFD = new File( latFilename );

	    // begin Mehmet
	    cacheFileName = new String( baseDir + "Cache.ser");
	    File cacheFD = new File( cacheFileName );
	    respTimeFileName = new String( baseDir + "RespTimes.ser");
	    File respTimeFD = new File( respTimeFileName );
	    // end Mehmet

            if( !latFD.exists() ){
                // create a new lat table
                lat = new LatTable();
            } else {
                // deserialize an existing lat table
                FileInputStream in = new FileInputStream( latFD );
                ObjectInputStream oin = new ObjectInputStream( in );
                lat = (LatTable) oin.readObject();
		oin.close();
            }

	    // begin Mehmet
	    if( !cacheFD.exists() ){
		// create a new cache index
		cache = new Cache();
	    } else {
		// deserialize an existing cache index
		FileInputStream in = new FileInputStream( cacheFD );
		ObjectInputStream oin = new ObjectInputStream( in );
		cache = (Cache) oin.readObject();
		oin.close();
		if (cache == null) {
		  cache = new Cache();
		}
	    }
	    if( !respTimeFD.exists() ){
		// create a new response time table
		respTimes = new RespTimeTable();
	    } else {
		// deserialize an existing response time table
		FileInputStream in = new FileInputStream( respTimeFD );
		ObjectInputStream oin = new ObjectInputStream( in );
		respTimes = (RespTimeTable) oin.readObject();
		oin.close();
		if (respTimes == null) {
		  respTimes = new RespTimeTable();
		}
		else hostServer = respTimes.getHostName();

		// Mehmet: the next line is put in non-browser tests in order
		// 		to enforce a HostServer selection at the
		//		beginning of the test. This line should
		//		be commented out if used in browser version.
		hostServer = null;

		// If host server is already known, inform that host server about that
		if (hostServer != null) {
          	  try {
          	    String result = doPostRequest("http://" + hostServer, "HostServer");
          	  } catch (Exception e) {
          	    System.out.println("Error: Cannot send HostServer msg to my host: " + hostServer);
		    writeToLogWithTime("NotResponding " + hostServer + " HostServer");
            	  }
        	}
	    }
	    // end Mehmet
        } catch( ClassNotFoundException e ){
            // create a new lat table
            lat = new LatTable();
	    cache = new Cache();		// Mehmet
	    respTimes = new RespTimeTable();	// Mehmet

            System.out.println( "init: lat table read exception: " +
                                e.getMessage() );
        }

        catch( IOException e ){
            // create a new lat table
            lat = new LatTable();
	    cache = new Cache();		// Mehmet
	    respTimes = new RespTimeTable();	// Mehmet

            if( Webpp.SERVER_CHOICE_DEBUG ){
                System.out.println( "init: lat table exception: " +
                                    e.getMessage() );
            }
        }

        catch( Exception e ){
            System.out.println( "init exception: " + e.getMessage() );
        }	

    }

    // applet destructor ======================================================
    public void old_destroy()
    {
	//System.out.println("In Webpp destructor");
        serLatTable();
	serCache();		// Mehmet: added this fcn call
	serRespTimesTable();	// Mehmet: added this fcn call
    }
    
    // Select a URL to send a request =========================================
    public String getUrl( String cArgs ) throws Exception
    {   
        Timer tApplet = null;     // measures time to execute the applet
        Timer tResource = null;   // measures time to receive resource from net
        Vector args, embRes;	  // Mehmet: added embRes declaration
	String embStr;		  // Mehmet: string to add to request for embedded resources
        Hashtable servers;
        String bestUrl = null;
        String bestServer = null;
        Stat1 stat = null;
        boolean first = true;
	        
        if( Webpp.timing ){
            tApplet = new Timer();
            tApplet.start();
            tResource = new Timer();
            tResource.start();
        }

	if (cArgs.indexOf("ServerList") >=0) {
	  System.out.println("Selecting host server...");
	  dummyFileName = baseDir + "dummy.html";
	  if (hostServer == null) findHostServer(null, this);	// Mehmet: determine the host server by sending dummy requests
	  else selectHostServer(null);	// Mehmet: determine the host server by simply choosing the best server from the latency table (after receiving an up to date server list from the NAME_SERVER)
	  return Net.patchPath(dummyFileName);
	}

	//System.out.println("Mehmet: cArgs=" + cArgs);	// Mehmet: debug
        args = getArgs( cArgs );
	servers = getServers( args );
        // Mehmet: added the following two lines to extract the list
        // of embedded resources from the URL (arguments of applet),
 	// and to check which ones are already in client cache.
        embRes = getEmbeddedResources( args );
	if (useCache)
	  embStr = cache.checkEmbResources(embRes);
	else embStr = null;
	//System.out.println("Mehmet: embStr=" + embStr);	// Mehmet: debug

        while( (bestServer = lat.getBestServer(servers)) != null ){
	    //System.out.println("Mehmet: bestServer: " + bestServer);	// Mehmet: debug
            Assert.assert( bestServer != null );

            if( Webpp.SERVER_CHOICE_DEBUG ){
                System.out.println( "best server: " + bestServer );
                System.out.println( "lat. table: " );
                lat.print();
            }
            
            if( Webpp.trace ){
                System.out.println( "selected server: " + bestServer );
            }
            
            bestUrl = (String) servers.get( bestServer );
	    //System.out.println("Mehmet: bestUrl: " + bestUrl);	// Mehmet: debug
            Assert.assert( bestUrl != null );

            if( first && Webpp.timing ){
                tApplet.stop();
                //System.out.println( "time for getUrl() exec = " + tApplet.getTime() );
                first = false;
            }

	    // Begin Mehmet: get previous resp time, if there is one
	    //	Important note: this has to be done before any call to Net.getDoc() since it is used there.
	    if (respTimes.containsKey(bestServer))
	      prevResponseTime = (String) respTimes.get(bestServer);
	    else
	      prevResponseTime = null;
	    // End Mehmet

            //showStatus( "Contacting " + bestServer + "..." );
	    //System.out.println("Contacting " + bestServer + "...");

	    // Mehmet: added if-else statements here
	    if (embStr != null)	
                stat = Net.getDoc( bestUrl + embStr, this); 
	    else
	        stat = Net.getDoc( bestUrl, this );
            
            if( Webpp.timing ){
                tResource.stop();
                System.out.println( "Response Time : " + 
                                    tResource.getTime() );
		responseTime = tResource.getTime();	// Mehmet: response time
		if( stat != null ) {
		  logStr += " " + bestServer;
		  logStr += " " + Long.toString(responseTime);
		}
            }

            if( stat != null ){
	        // successfully retrieved the document
                //showStatus( "Contacted " + bestServer );
	        //System.out.println("Contacted " + bestServer);
	        
	        if( Webpp.SERVER_CHOICE_DEBUG ){
                    System.out.println( "rec OK, latency = " + stat.lat );
                }
            
		// Update latency table
                lat.addLatency( bestServer, stat.lat );

		// Begin Mehmet: insert response time to the respTimes table
		if (responseTime > 0) {
		  String fnameStr = new String (bestUrl);
		  // Take the only filename portion of the URL if possible (i.e. unless URL ends with '/')
		  int separatorPos = bestUrl.lastIndexOf('/');
	  	  if (separatorPos != -1 && separatorPos < (bestUrl.length() - 1))
		    fnameStr = bestUrl.substring(separatorPos + 1);

		  // Mehmet: use the next line if you want to piggyback response time.
		  //	Otherwise, use the line following it in order to piggyback the latency.
		  //String rtime = (new Long(responseTime)).toString();
		  String rtime = (new Long((long) (stat.lat * 1000))).toString();

		  respTimes.put(bestServer, fnameStr + "|" + rtime);
		}
		// End Mehmet
                break;
                
            } else {
                // mark server as unreachable
                //showStatus( "Could not reach " + bestServer );
	        //System.out.println("Could not reach " + bestServer);
                
		//// Mark as unreachable only if this is not the NAME_SERVER
		//if (!NAME_SERVER.equals(bestServer) &&
		//	!NAME_SERVER.startsWith(bestServer) && 
		//	!NAME_SERVER_IP.equals(bestServer) && 
		//	!NAME_SERVER_IP.startsWith(bestServer))
		//  lat.updateReachability( bestServer, false );
	
		// Mark the server as unreachable
                lat.updateReachability( bestServer, false );
		writeToLogWithTime("Unreachable " + bestServer + " ClientRequest");
                
                if( Webpp.trace ){
                    System.out.println( "server " + bestServer + 
                                        " unreachable" );
                }
            }
	}

        // asynchronously write lat table to disk
        // guarantees no lost updates for next invokation of the applet
        // (as there is no guarantee that destroy() gets called beforehand)
        chptLatTable( this );
	
	chptCache( this );	// Mehmet: asynchronously write cache index to disk

	chptRespTimes( this );	// Mehmet: asynchronously write respTimes table to disk
        
        if( stat != null ){
            return stat.locURL;
        } else {
            // failure to find the requested document should be 
            // handled more gracefully
            return null;
        }
    }	// getUrl()
    
    
    // serialize lat table onto disk =======================================
    void serLatTable()
    {
        FileOutputStream out;
        ObjectOutputStream oout;
        
        // serialize current latency table
        Assert.assert( latFilename != null );

        try {
            out  = new FileOutputStream( latFilename );
            oout = new ObjectOutputStream( out );

            oout.writeObject( lat );
            oout.flush();
            oout.close();
        }

        catch( IOException e ){
            System.out.println( "destroy exception: " + e.getMessage() );
        }
    }


    // Mehmet: serialize cache onto disk =======================================
    void serCache()
    {
        FileOutputStream out = null;
        ObjectOutputStream oout = null;
        
        // serialize current cache index
        Assert.assert( cacheFileName != null );

        try {
	    if (cache != null && cache.size() > 0) {
              out  = new FileOutputStream( cacheFileName );
	      if (out != null)
                oout = new ObjectOutputStream( out );
	    }

	    if (oout != null) {
              oout.writeObject( cache );
              oout.flush();
              oout.close();
	    }
        }

        catch( IOException e ){
            System.out.println( "serCache exception: " + e.getMessage() );
        }
    }


    // Mehmet: serialize respTimes table onto disk =======================================
    void serRespTimesTable()
    {
        FileOutputStream out = null;
        ObjectOutputStream oout = null;
        
        // serialize current response times table
        Assert.assert( respTimeFileName != null );

        try {
	    if (respTimes != null && respTimes.size() > 0) {
              out  = new FileOutputStream( respTimeFileName );
	      if (out != null)
                oout = new ObjectOutputStream( out );
	    }

	    if (oout != null) {
              oout.writeObject( respTimes );
              oout.flush();
              oout.close();
	    }
        }

        catch( IOException e ){
            System.out.println( "serRespTimesTable exception: " + e.getMessage() );
        }
    }


    // asynchronously serialize latency table to disk =======================
    private void chptLatTable( Webpp self )
    {
        new ChptLatTable( self ).run();
    }


    // Mehmet: asynchronously serialize cache index to disk =======================
    private void chptCache( Webpp self )
    {
        new ChptCache( self ).run();
    }


    // Mehmet: asynchronously serialize respTimes table to disk =======================
    private void chptRespTimes( Webpp self )
    {
        new ChptRespTimes( self ).run();
    }

// Mehmet: initiate the procedure for finding the closest (host) server
//	to this client. Then, host server is notified by a POST request.
    public void chooseHostServer()
    {
	try {
          getUrl("ServerList");
        } catch ( Exception e ) {
          System.out.println( "Exception in doGet while trying to choose host server: " + e);
        }
    }

    /* selectHostServer: Determines the host server by simply choosing
     * the best server from the latency table (after receiving an
     * up to date server list from the NAME_SERVER, it uses that
     * list to also insert the new servers into the latency table).
     */
    public void selectHostServer( String serverNames )
    {
	String url, bestServer;
	System.out.println("Choosing host server from latency table entries");
        if (serverNames == null) {
          url = new String("http://" + NAME_SERVER);
          //System.out.println("URL: " + url);
          try {
            serverNames = doPostRequest(url, "ServerList");
          } catch (Exception e) {
            System.out.println("Inside serverNames (selectHostServer): " + serverNames);
            serverNames = null;
            System.out.println("Error: cannot send ServerList request to NAME_SERVER: " + e);
          }
          // If name server cannot be contacted, also try its IP address
          if (serverNames == null) {
            url = new String("http://" + NAME_SERVER_IP);
            //System.out.println("URL: " + url);
            try {
              serverNames = doPostRequest(url, "ServerList");
            } catch (Exception e) {
              System.out.println("Inside2 serverNames (selectHostServer): " + serverNames);
              serverNames = null;
              System.out.println("Error: cannot send ServerList request to NAME_SERVER: " + e);
              writeToLogWithTime("NotResponding " + NAME_SERVER + " ServerList");
            }
          }
          //System.out.println("Received ServerList in selectHostServer is : " + serverNames);
        }
        if (serverNames == null) {
          System.out.println("Error: ServerList is empty in selectHostServer()");
          return;
        }
	Hashtable servers = new Hashtable();
	StringTokenizer st = new StringTokenizer(serverNames, " ");
	while (st.hasMoreTokens()) {
          url = (String) st.nextToken();	// url includes the host name and port number
	  servers.put(url, url + "/");
	}
	try {
	  bestServer = lat.getBestServer(servers);
	} catch (Exception ee) { bestServer = null; }
        if (bestServer != null) {
	  hostServer = bestServer;
	  System.out.println("Host server chosen from latency table: " + hostServer);

	  // Finally, send a POST request to the closest server to let it know
          // about that.

          try {
            String result = doPostRequest("http://" + hostServer, "HostServer");
          } catch (Exception e) {
            System.out.println("Error: Cannot send HostServer msg to my host: " + e);
            writeToLogWithTime("NotResponding " + hostServer + " HostServer");
	    findHostServer(serverNames, this);	// enforce selection of host server by sending dummy requests
          }
          writeToLog(Long.toString(System.currentTimeMillis() - testStartTime) + " Client " + server + " HostServer " + hostServer + "\n");
        }
	else findHostServer(serverNames, this);	// enforce selection of host server by sending dummy requests
    }	// selectHostServer()

    // Mehmet: finds the closest (host) server to this client
    //  by receiving the list of active servers from the
    //  NAME_SERVER and then sending dummy requests to all active servers.
    public void findHostServer(String serverNames, Webpp self)
    {
	dummyRequestsStarted = true;
	dummyRequestsFinished = false;
	HostServerFinder hsf = new HostServerFinder( self, serverNames );
	hsf.runit();
    }

    // Unpack arguments from a string =======================================
    private Vector getArgs( String cArgs )
    {
        StringTokenizer st = new StringTokenizer( cArgs, " " );
        Vector argv = new Vector();

        for (; st.hasMoreTokens(); argv.addElement( st.nextToken() ) );

        return argv;        
    }

    // Get the embedded resources list out of the args vector
    private Vector getEmbeddedResources( Vector args )
    {
      Enumeration e = args.elements();
      String embRes = (String) e.nextElement();
      StringTokenizer st = new StringTokenizer(embRes, ";");
      Vector embResVec = new Vector();
      for (; st.hasMoreTokens(); embResVec.addElement( st.nextToken() ) );

      return embResVec;
    }

    // Filter out a list of servers from a list of URLs =======================
    //    (the hashtable maps each server to the corresponding URL)
    private Hashtable getServers( Vector urls )
    {
        Hashtable servers = new Hashtable();

        try {
            String urlS;
            URL url = null;
            String host;
            int    port;
            String server;

            Enumeration e = urls.elements();
            urlS = (String) e.nextElement();
            for ( ; e.hasMoreElements(); ) {

                urlS = (String) e.nextElement();
		if (urlS == null) continue;
		//System.out.println("Mehmet: next element=" + urlS);	// Mehmet: debug
		try {
		  url  = new URL( urlS );
		} catch (MalformedURLException er) {
		  //System.out.println("Error in getServers: MalformedURLException for url: " + urlS); 
		  System.out.println("Error in getServers: MalformedURLException");
		  url = null;
		}
		if (url == null) continue;
                host = url.getHost();
                port = url.getPort();

                if( port == -1 ){
                    server = host;
                } else {
                    server = host + ":" + Integer.toString( port );
                }

                servers.put( server, urlS );
            }
        }

        catch( Exception e1 ){
	        //System.out.println( "Error: exception in getServers: " + e1.getMessage() );
		System.out.println( "Error: exception in getServers!");
        }

        return servers;
    }

    // Get fastest server =====================================================
    private String getBestUrl( Hashtable servers )
        throws Exception
    {
        return lat.getBestServer( servers );
    }
    
    //// write a string to Java console (used for instrumenting HTML) ===========
    //public void log( String msg )
    //{
    //    System.out.println( msg );
    //}

    // Mehmet: doPostRequest() =============================================
    //  performs a POST request for the given request type to the given server.
    public String doPostRequest(String urlName, String reqType) throws Exception {
      long bytesRead;
      byte b = 0;
      int done = 0;

      String str = new String(urlName + "/servlet/webppserv.Webpp/?" + reqType);
      URL url = null;
      try {
	url = new URL(str);
      } catch (MalformedURLException e) {
        System.out.println("Error in doPostRequest: MalformedURLException for url: " + url);
      }
      //System.out.println("DoPost: " + url);
      //HttpURLConnection conn = null;
      URLConnection conn = null;
      try {
        //conn = (HttpURLConnection) url.openConnection();
        conn = url.openConnection();
      } catch (Exception e) {
        System.out.println("Error in doPostRequest openConnection: " +
                                                e.getMessage() );
        return null;
      }

      //System.out.println("HERE: doPostRequest");

//      conn.setDoOutput(true);
//      try {
//        conn.setRequestMethod("POST");
//      } catch (ProtocolException pe) {
//      System.out.println("Error: cannot set reqType!");
//      }

      String result = null;
      StringBuffer buf = new StringBuffer();
      //PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter("out.txt")));
      bytesRead = 0;

      //DataOutputStream outs = new DataOutputStream(conn.getOutputStream());
      //outs.writeUTF(reqType);
      //outs.close();

      InputStream instream = conn.getInputStream();
      if (instream == null) return null;
      DataInputStream in = new DataInputStream(instream);
      if (in == null) return null;
      try { b = in.readByte(); } catch (EOFException e) { done = 1; }
      while (done == 0) {
        // Mehmet: comment out the next three lines if you don't want to echo the downloaded file to screen
        //System.out.print((char)b);
        //if (b==13 || b==10) out.println("");
        //else out.print((char)b);
        buf.append((char) b);
        bytesRead++;
        try { b = in.readByte(); } catch (EOFException e) { done = 1; }
      }
      //out.close();
      in.close();

      if (buf.length() > 0)
        result = buf.toString();

      if (bytesRead > 0) return result;
      else return null;
    }

    // main ===================================================================
    public void main( String[] argv )		// Webpp class
    {
        String url;

        if ( MAIN_DEBUG ) {
            try {           
	            for (int i = 0; i < 2; i++) {
                    System.out.println( "iteration " + i );
                
                    url = getUrl( "http://coke.stanford.edu:8088/w05K.txt" +
                         " " + "http://frankfurt.cs.uni-sb.de:8088/w05K.txt" );

                    Thread.currentThread().sleep( 4000 );
                }            
            }
            
            catch ( Exception e ) {
                System.out.println( "Exception in main: " + e.getMessage() );
                Runtime.getRuntime().exit( -1 );
            }                     
        }   
    }

    // ====================  doGet  ========================
    protected void doGet( HttpServletRequest req, HttpServletResponse resp )
                                                throws ServletException, IOException
    {
        BufferedReader reader = null;
        ServletOutputStream out = null;
        BufferedInputStream in = null;
        String mimeType = null;
        byte buf[] = new byte[BUFFER_SIZE];
        int bytesRead;

	String reqString = req.getQueryString();
	if ( (reqString != null) && (reqString.indexOf("TestPhase") >=0) ) {
          doPost(req, resp);
          return;
        }

	if( server == null ) {
	  server = req.getServerName();
	  initWebppClient();	// Initialize client
	}

    }	// doGet()

    // ====================  doPut  ========================
    protected void doPut( HttpServletRequest req, HttpServletResponse resp )
						throws ServletException, IOException
    {
        ServletInputStream reader = null;
        String storeFile = null;
        FileOutputStream store = null;
        String line = null;
        String qry = null;
        Vector pars;
        String urn = null;
        String urlList = null; 
        byte buffer[] = new byte[4096];
        int len = 0;   

        //TODO: copy and paste parts of doGet from servlet code
    }

    // ====================  doPost  ========================
    protected void doPost( HttpServletRequest req, HttpServletResponse resp )
						throws ServletException, IOException
    {
	ServletOutputStream out = null;
	String qry = null, reqType = null, parameters = null;
        Vector pars;
	int len = 0;
	byte buffer[] = new byte[4096];   
	boolean calledInitInThisRequest = false;

	calledInitInThisRequest = false;
        if( server == null ) {
          server = req.getServerName();
	  initWebppClient();	// Initialize client
	  calledInitInThisRequest = true;
        }

	qry = req.getQueryString();
        if( qry == null ) {
          resp.sendError( HttpServletResponse.SC_NOT_FOUND,
                                        "doPost: no query string received" );
          return;
        }

        System.out.println("\nPost query str: " + qry);

	// parse qry string
	StringTokenizer st = new StringTokenizer(qry, "&");
	for (int i=0; st.hasMoreElements(); i++) {
	  if( i == 0 ){
            reqType = ((String) st.nextElement());
	  } else if( i == 1 ){
	    parameters = ((String) st.nextElement());
	  } else {
	    parameters += " " + ((String) st.nextElement());
	  }
	}

	if (reqType == null) {
          resp.sendError( HttpServletResponse.SC_NOT_FOUND,
                                        "doPost: no request type specified");
          return;
        }

        out = resp.getOutputStream();

	if (reqType.equalsIgnoreCase("TestPhase")) {
	  String tmp = parameters;      // the remainder of parameters is only the new test phase number
          testPhase = Integer.valueOf(tmp).intValue();
	  if (testPhase == 0 && !calledInitInThisRequest)
	    initWebppClient();	// Initialize client
	  System.out.println("Test Phase: " + tmp);
          writeToLog(Long.toString(System.currentTimeMillis() - testStartTime) + " Client " + server + " TestPhase " + tmp + "\n");
	}	// TestPhase
    }	// doPost()

   /*
    *=========================================================================
    * Write the given string to the log file after adding standard log
    *  info at the beginning of it (e.g. current test time and client name)
    *=========================================================================
    */
    static void writeToLogWithTime(String str)
    {
      if (str == null) {
        System.out.println("Error: str is null in writeToLog");
        return;
      }
      str = Long.toString(System.currentTimeMillis() - testStartTime) + " Client " + server + " " + str + "\n";
      writeToLog(str);
      //try {
      //  logfile.writeBytes(str);
      //} catch (IOException e) {
      //  System.out.println("Error: cannot write to logfile: " + e);
      //}
    }

   /*
    *=========================================================================
    * Write the given string to the log file
    *=========================================================================
   */
    static synchronized void writeToLog(String str)
    {
	if (str == null) {
	  System.out.println("Error: str is null in writeToLog");
	  return;
	}

	try {
	  logfile.writeBytes(str);
        } catch (IOException e) {
          System.out.println("Error: cannot write to logfile: " + e);
        }
    }

   /*
    *=========================================================================
    * Get the current test time from the NAME_SERVER
    *=========================================================================
   */
    public long getTestTimeFromNameServer()
    {
        long result = 0, tmplong;
	String tmp = null;

	String url = new String("http://" + NAME_SERVER);
        try {
          tmp = doPostRequest(url, "GetTestTime");
        } catch (Exception e) {
          tmp = null;
          System.out.println("Error: cannot send GetTestTime request to NAME_SERVER: " + e);
        }
        // If name server cannot be contacted, also try its IP address
        if (tmp == null) {
          url = new String("http://" + NAME_SERVER_IP);
          try {
            tmp = doPostRequest(url, "GetTestTime");
          } catch (Exception e) {
            tmp = null;
            System.out.println("Error: cannot send GetTestTime request to NAME_SERVER: " + e);
	    writeToLogWithTime("NotResponding " + NAME_SERVER + " GetTestTime");
          }
	}

	if (tmp == null) return 0;
	tmplong = (new Long(tmp)).longValue();
	System.out.println("Received test time: " + tmp);
	result = System.currentTimeMillis() - tmplong;
        return result;
    }	// getTestTimeFromNameServer()

}	// Webpp class

///////////////////////////////////////////////////////////////////////////////
// 
// Thread for asynchronous lat table checkpointing
//
///////////////////////////////////////////////////////////////////////////////
final class ChptLatTable extends Thread
{
    // constructor ============================================================
    ChptLatTable( Webpp app )
    {
        // remember reference to parent applet
        applet_ = app;
    }
    
    // asynchronously writes lat table to disk =============================
    public void run()
    {          
        applet_.serLatTable();
    }

    // private data members ===================================================
    private Webpp applet_;
}


///////////////////////////////////////////////////////////////////////////////
// 
// Mehmet: Thread for asynchronous cache index checkpointing
//
///////////////////////////////////////////////////////////////////////////////
final class ChptCache extends Thread
{
    // constructor ============================================================
    ChptCache( Webpp app )
    {
        // remember reference to parent applet
        applet_ = app;
    }
    
    // asynchronously writes cache index to disk =============================
    public void run()
    {          
        applet_.serCache();
    }

    // private data members ===================================================
    private Webpp applet_;
}


///////////////////////////////////////////////////////////////////////////////
// 
// Mehmet: Thread for asynchronous response times table serialization
//
///////////////////////////////////////////////////////////////////////////////
final class ChptRespTimes extends Thread
{
    // constructor ============================================================
    ChptRespTimes( Webpp app )
    {
        // remember reference to parent applet
        applet_ = app;
    }
    
    // asynchronously writes response times table to disk =============================
    public void run()
    {          
        applet_.serRespTimesTable();
    }

    // private data members ===================================================
    private Webpp applet_;
}


///////////////////////////////////////////////////////////////////////////////
// 
// Mehmet: Thread for determining the closest (host) server
//
///////////////////////////////////////////////////////////////////////////////

final class HostServerFinder extends Thread
{
    // constructor ============================================================
    HostServerFinder( Webpp app, String serverNames )
    {
	//System.out.println("HERE: In HostServerFinder");
        // remember reference to parent applet
        applet_ = app;

	// get serverNames string from the argument
	this.serverNames = serverNames;
    }

    // Mehmet: getDummyURL() ================================================
    //	performs a get request to the dummy.html at the
    //	given server to measure its latency. This function is called when
    //	the Web++ client receives the list of server names from a servlet
    //	and starts measuring their latencies to pick the closest server
    //	(host server) to itself.
    public long getDummyURL(String urlName) throws Exception {
      long first_byte_time, bytesRead;
      long time1, time2=0;
      byte b = 0;
      int done = 0;

      URL url = null;
      try {
	url = new URL("http://" + urlName + "/dummy.html");
      } catch (MalformedURLException e) {
        System.out.println("Error in getDummyURL: MalformedURLException for url: " + url);
      }
      URLConnection conn = url.openConnection();
      //PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter("out.txt")));
    
      bytesRead = 0;
      time1 = System.currentTimeMillis();
      InputStream instream = conn.getInputStream();
      if (instream == null) return 0;
      DataInputStream in = new DataInputStream(instream);
      try { b = in.readByte(); } catch (EOFException e) { done = 1; }
      while (done == 0) {
        // Mehmet: comment out the next three lines if you don't want to echo the downloaded file to screen
        //System.out.print((char)b);
        //if (b==13 || b==10) out.println("");
        //else out.print((char)b);
        bytesRead++;
        if (bytesRead == 1) { time2 = System.currentTimeMillis(); }
        try { b = in.readByte(); } catch (EOFException e) { done = 1; }
      }
      //out.close();
      in.close();
    
      first_byte_time = time2 - time1;
      if (bytesRead > 0 && time2 > 0) return first_byte_time;
      else return 0;
    }


    // Mehmet: doPostRequest() =============================================
    //	performs a POST request for the given request type to the given server.
    public String doPostRequest(String urlName, String reqType) throws Exception
    {
      long bytesRead;
      byte b = 0;
      int done = 0;

      //yield();
      //System.out.println("Before Sleep: " + System.currentTimeMillis());
      //sleep(10000);
      //System.out.println("After  Sleep: " + System.currentTimeMillis());

      String str = new String(urlName + "/servlet/webppserv.Webpp/?" + reqType);
      URL url = null;
      try {
	url = new URL(str);
      } catch (MalformedURLException e) {
        System.out.println("Error in doPostRequest: MalformedURLException for url: " + url);
      }
      //System.out.println("DoPost: " + url);
      //HttpURLConnection conn = null;
      URLConnection conn = null;
      try {
	//conn = (HttpURLConnection) url.openConnection();
	conn = url.openConnection();
      } catch (Exception e) {
	System.out.println("Error in doPostRequest openConnection: " +
						e.getMessage() );
	return null;
      }

      String result = null;
      StringBuffer buf = new StringBuffer();
      //PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter("out.txt")));
      bytesRead = 0;
    
      InputStream instream = conn.getInputStream();
      if (instream == null) return null;
      DataInputStream in = new DataInputStream(instream);
      try { b = in.readByte(); } catch (EOFException e) { done = 1; }
      while (done == 0) {
        // Mehmet: comment out the next three lines if you don't want to echo the downloaded file to screen
        //System.out.print((char)b);
        //if (b==13 || b==10) out.println("");
        //else out.print((char)b);
	buf.append((char) b);
        bytesRead++;
        try { b = in.readByte(); } catch (EOFException e) { done = 1; }
      }
      //out.close();
      in.close();

      if (buf.length() > 0)
	result = buf.toString();
    
      if (bytesRead > 0) return result;
      else return null;
    }
    
    //public void run()
    public void runit()
    {          
	String bestServer = null, url = null;
	long bestLatency = 0;
	FileOutputStream fos = null;
	//System.out.println("HERE: In runit");
	try {
	  fos = new FileOutputStream(applet_.dummyFileName);
	} catch (IOException ioe) {
	  System.out.println("Error: cannot create " + applet_.dummyFileName);
	  return;
	}
	PrintWriter dummyOut = new PrintWriter(fos);

	//System.out.println("Choosing the host server...<BR>");
	//System.out.println("Prev serverNames: " + serverNames);
	if (dummyOut != null)
	  dummyOut.println("Choosing the host server...<BR>");

	//// Wait 120 seconds before doing anything
	//while ((System.currentTimeMillis() - applet_.testStartTime) < 120000);
	
	//yield();

	// If serverNames string is empty, then request the serverNames list from
	// the name server (coordinator) by sending a POST request.
	if (serverNames == null) {
	  url = new String("http://" + applet_.NAME_SERVER);
	  //System.out.println("URL: " + url);
	  try {
	    serverNames = doPostRequest(url, "ServerList");
	  } catch (Exception e) {
	    System.out.println("Inside serverNames: " + serverNames);
	    serverNames = null; 
	    System.out.println("Error: cannot send ServerList request to NAME_SERVER: " + e);
	  }
	  // If name server cannot be contacted, also try its IP address
	  if (serverNames == null) {
            url = new String("http://" + applet_.NAME_SERVER_IP);
	    //System.out.println("URL: " + url);
            try {
              serverNames = doPostRequest(url, "ServerList");
            } catch (Exception e) {
	      System.out.println("Inside2 serverNames: " + serverNames);
	      serverNames = null; 
	      System.out.println("Error: cannot send ServerList request to NAME_SERVER: " + e);
	      applet_.writeToLogWithTime("NotResponding " + applet_.NAME_SERVER + " ServerList");
	    }
	  }
	  //System.out.println("Received ServerList is : " + serverNames);
	  if (dummyOut != null)
	    dummyOut.println("Received ServerList is : " + serverNames + "<BR>");
	}

	if (serverNames == null) {
	  System.out.println("Error: ServerList is empty in HostServerFinder::run()");
	  if (dummyOut != null) {
	    dummyOut.println("Error: cannot get the server list!<BR>");
	    dummyOut.close();
	  }
	  return;
	}

        // First, send GET requests to all servers and measure latencies.
	// The function getDummyURL() is used for requesting the dummy.html
	// at each server.
	StringTokenizer st;
	long latency = 0;
	st = new StringTokenizer(serverNames, " ");
	while (st.hasMoreTokens()) {
	  url = (String) st.nextToken();
	  if (url == null) continue;
	  try {
	    latency = getDummyURL(url);
	  } catch (Exception e) { latency = 0; }
	  if (latency > 0) {
	    // The next line inserts the latency to server latency table so that 
	    //	it can be used for later bestURL selections
	    //applet_.lat.insertLatency( url, (new Long(latency)).doubleValue() / 1000.0);
	    //System.out.println("DummyRequest URL: " + url + "  Latency: " + latency);
	    if (dummyOut != null)
	      dummyOut.println("URL: " + url + " Latency: " + latency + "<BR>");
	    if (bestLatency == 0 || latency < bestLatency) {
	      bestLatency = latency;
	      bestServer = url;
	    }
	  }
	  else {
	    //System.out.println("DummyRequest no reply from " + url);
	    if (dummyOut != null)
	      dummyOut.println("No reply from " + url + "<BR>");
	  }
	}
	
	if (bestServer == null) {
	  System.out.println("Error in HostServerFinder::run() : cannot find a host server!");
	  if (dummyOut != null) {
	    dummyOut.println("Error: cannot find a host server!<BR>");
	    dummyOut.close();
	  }
	  //yield();
	  return;
	}

	// Next, set the static variable hostServer to the name of the closest
	// server in Webpp class.

	applet_.hostServer = bestServer;
	applet_.dummyRequestsFinished = true;
	applet_.dummyRequestsStarted = false;
	applet_.respTimes.putHostName(applet_.hostServer);
	System.out.println("Host server chosen: " + applet_.hostServer);
	if (dummyOut != null)
	  dummyOut.println("Host server chosen: " + applet_.hostServer + "<BR>");

	//yield();

	// Finally, send a POST request to the closest server to let it know
	// about that.

	if (applet_.hostServer != null) {
	  try {
	    String result = doPostRequest("http://" + applet_.hostServer, "HostServer");
	  } catch (Exception e) {
	    System.out.println("Error: Cannot send HostServer msg to my host: " + e);
	    applet_.writeToLogWithTime("NotResponding " + applet_.hostServer + " HostServer");
	  }
          applet_.writeToLog(Long.toString(System.currentTimeMillis() - applet_.testStartTime) + " Client " + applet_.server + " HostServer " + applet_.hostServer + "\n");
	}

	if (dummyOut != null)
	  dummyOut.close();	// close dummy file so that browser can display it
    }

    // private data members ===================================================
    private Webpp applet_;
    private String serverNames;

}

