package webppserv;

import java.net.*;
import java.io.*;
import java.util.*;
import javax.servlet.*;
import com.oroinc.text.regex.*;
import java.util.zip.*;		// Mehmet: added for checksum calculation
import java.lang.*;		// Mehmet: added for checksum calculation


///////////////////////////////////////////////////////////////////////////////
// 
// IO filter which does HTML HREF and SRC tag substitution.
//
///////////////////////////////////////////////////////////////////////////////
class BatchInputStream extends BufferedInputStream
{
   /*
    *=========================================================================
    * private class variables
    *=========================================================================
    */
    
    /* regex compiler machinery */ 
    static PatternMatcher matcher;
    static PatternCompiler compiler;
    protected static Pattern pattern[];
    protected static Pattern lineSepPattern;
    protected static Pattern bodyPattern;
    protected static Pattern srcPattern;
    
    /* substitution patterns */ 
    protected static String rule[];
    protected static String prefix;
    protected static String suffix;
       
    /* stream with parent doc */
    protected ByteArrayInputStream parentDocStream = null;
    
    /* list of embedded resources */
    protected Hashtable embFileNames = new Hashtable();

    /* reference to a parent servlet */
    protected Webpp servlet;
    
    /* directory of parent html document */
    protected String parentDocDir;
    
    /* state of read() */
    private BufferedInputStream currentEmbStream = null;
    private Enumeration embDocEnum = null;
    private boolean parentDoc = false;


   /*
    *=========================================================================
    * object constructor
    *=========================================================================
    */
    protected BatchInputStream( String fileName, Webpp parentServlet ) 
        throws FileNotFoundException
    { 
        super( new FileInputStream(fileName) );
        
        String line = null;
        StringBuffer buf1 = new StringBuffer( Webpp.BUFFER_SIZE );
        String buf2 = null;
        String replace = null;
        String sep;
        String fileSep;
        String replList = null;
        PatternMatcherInput input = null;
        MatchResult result = null;
        int lastEnd = 0;
        String embFileName = null;
        BufferedReader parentDoc = null;
        String docType;
        boolean isHtml;
	long t0, t1 = 0, tt;		// Mehmet: measure times
        
        
        servlet = parentServlet;

        docType = servlet.getServletConfig().getServletContext().
                  getMimeType( fileName );
        isHtml = ( docType != null && docType.indexOf( "html" ) != -1 );
        
        sep = System.getProperty( "line.separator" );
        fileSep = System.getProperty( "file.separator" );
 
        parentDoc = new BufferedReader(new FileReader(fileName));
        parentDocDir = fileName.substring(0, fileName.lastIndexOf(fileSep) + 1);

        try {
            while( (line = parentDoc.readLine()) != null ){
                if( buf1 == null ){
                    buf1.append( line );
                } else {
                    buf1.append( sep );
                    buf1.append( line );
                }
            }
        } catch( IOException e ){
            Assert.assert( false, e.getMessage() + "when reading " + 
                                  fileName );
        }
        
	if( isHtml ){
	    buf2 = buf1.toString();
	}

	// parse all SRC and generate a vector with file names 
        if( isHtml ){
            input = new PatternMatcherInput( buf2 );
	    lastEnd = 0;	// Mehmet: for checksum substitution
            while( matcher.contains(input, srcPattern) ){
                result = matcher.getMatch();
                embFileName = parentDocDir + result.group(1);
		embFileNames.put( embFileName, new String( " " ) );
	    }

            embDocEnum = embFileNames.keys();
        }
        
        try {
            parentDoc.close();
        } catch( IOException e ){
            Assert.assert( false, "closing " + fileName );
        } 
        if( isHtml ){
            parentDocStream = new ByteArrayInputStream( buf2.getBytes() ); 
        } else {
            parentDocStream = new ByteArrayInputStream( 
                                              buf1.toString().getBytes() );
        }
        
    }


   /*
    *=========================================================================
    * overloaded - iterates over the parent doc file and files of all emb docs
    *=========================================================================
    */
    public int read( byte buf[], int pos, int len) throws IOException
    {  
        int bytesRead;
        String embDocFileName;
        String separator;
        String docName;
        byte tmpBuf[];
        

        /* first send embedded resources */
        if( currentEmbStream != null &&
            (bytesRead = currentEmbStream.read( buf, pos, len )) != -1 ){
            return bytesRead;
        }
         
        /* open another emb resource file */
        while( embDocEnum != null && embDocEnum.hasMoreElements() ){
            embDocFileName = (String) embDocEnum.nextElement();
            
            if( currentEmbStream != null ){
                currentEmbStream.close();
            }
            try {
                currentEmbStream = new BufferedInputStream( 
                                      new FileInputStream( embDocFileName ) );
            } catch( FileNotFoundException e ){
                System.out.println( "emb. resource " + embDocFileName +
                                    " not found" );
                continue;
            }
            
            /* prepare a separator to be put between the files */
	    docName = embDocFileName.substring( parentDocDir.length(), 
						embDocFileName.length() );

	    /* Put the separator between the files */
            separator = "<WEBPP_APPEND " + docName + ">";
            tmpBuf = separator.getBytes();
            Assert.assert( tmpBuf.length > 0 );
            System.arraycopy( tmpBuf, 0, buf, 0, tmpBuf.length );
            return tmpBuf.length;
        }
        
        /* add separator for parent document */
        if( !parentDoc ){
            Assert.assert( servlet.reqFile != null );
            docName = servlet.reqFile.substring( Webpp.rootDir.length(),
                                                 servlet.reqFile.length() );
            separator = "<WEBPP_APPEND " + docName + ">";
            tmpBuf = separator.getBytes();
            Assert.assert( tmpBuf.length > 0 );
            System.arraycopy( tmpBuf, 0, buf, 0, tmpBuf.length );
                        
            parentDoc = true;

            return tmpBuf.length;
        }
        
        /* read while you can from parent document */
        bytesRead = parentDocStream.read( buf, pos, len );
        if( bytesRead != -1 ){            
            return bytesRead;
        }
        
        /* everything has been read */
        if( currentEmbStream != null ){
            currentEmbStream.close();
        }

        return -1;
    }
                   

   /*
    *=========================================================================
    * class constructor
    *=========================================================================
    */
    static 
    {
        int i = -1;

        /* initialize regex compiler */
        matcher = new Perl5Matcher();
        compiler = new Perl5Compiler();

        /* initialize substitution rules */
        rule = new String[] { "HREF\\s*=\\s*\"(\\S+)\"",
                              "HREF=\"\" onClick=\"this.href = " };
        prefix = "document.Webpp.getUrl(\'";
        suffix = "\')\"";

        pattern = new Pattern[rule.length/2];

        /* parse regular expressions */
        try {
            for( i = 0; i < rule.length; i += 2 ){
                pattern[i/2] = compiler.compile( rule[i],
                                       Perl5Compiler.CASE_INSENSITIVE_MASK );  
            }
            lineSepPattern = compiler.compile( "\n+" );
            bodyPattern    = compiler.compile( "<.*BODY.*>", 
                                       Perl5Compiler.CASE_INSENSITIVE_MASK );
            srcPattern = compiler.compile( "SRC\\s*=\\s*\"(\\S+)\"",
                                       Perl5Compiler.CASE_INSENSITIVE_MASK );
                                       
        } catch( Exception e ){
            Assert.assert( false, Integer.toString(i) + e.getMessage() );
        }
    }
}
