package webppserv;

import java.io.*;
import java.lang.*;
import java.util.*;
import java.net.*;

class ResourceDownloader extends Thread implements Runnable
{
  protected static final int HUGE_BUF_LEN = 307200;
  protected static final int BUF_LEN = 8192;
  protected static final String separator = new String( "<WEBPP_APPEND " );
  protected static final byte separatorB[] = separator.getBytes();

  String remoteServer, fname, url, localDir;

  public ResourceDownloader(String remoteServer, String fname, String url)
  {
	this.remoteServer = remoteServer;
	this.fname = fname;
	this.url = url;
	localDir = null;
  }

  public void run()
  {
	try { setPriority(MIN_PRIORITY); } catch (Exception ee) {}
	yield();	// give the caller a chance to go on

	// Figure out the directory path so that you can store the resource and
	//	its embedded resources in the same directory on the local disk
	int dirPos = url.toLowerCase().indexOf("webppserv.webpp");
	if (dirPos != -1) {
	  dirPos += 16;		// bypass "webppserv.Webpp" part of the URL
	  int endPos = url.lastIndexOf("/");
	  if (endPos == -1) endPos = url.lastIndexOf(System.getProperty("file.separator"));
	  if (endPos == -1 || endPos <= dirPos)
	    localDir = Webpp.rootDir;
	  else
	    localDir = Webpp.rootDir + url.substring(dirPos, endPos) +
						System.getProperty("file.separator");
	}     // end if (dirPos != -1)

	getDoc(url, localDir);

        System.out.println("Download is finished for " + fname);

	// Now, update the replication directory and broadcast an update
	//	message for the replication directory with the new replica location.
	int pos = url.indexOf(remoteServer);
	if (pos == -1) {
	  System.out.println("Error: cannot find the remote server name in the remote URL");
	  return;
	}
	int pos2 = url.indexOf('/', pos + 1);
	if (pos2 == -1) {
	  System.out.println("Error: cannot parse the remote URL");
	  return;
	}
	String newurl = "http://" + Webpp.server + ":" + Webpp.port + url.substring(pos2);
	Webpp.repl.appendUrl(fname, newurl);
	Webpp.broadcastMessage("AddToReplDir", fname + " " + newurl, false);
	yield();	// give a chance to broadcaster to start
  }

  // getDoc() : retrieves a given url and all of its embedded resources
  static void getDoc(String url, String localDir)
  {
	String locURL = null;
	URL doc;
	InputStream in;
	String cont;
	URLConnection conn = null;

        try {
            doc = new URL( url );
        } catch( Exception e ){
            //System.out.println( "getDoc: URL create " + e.getMessage() );
            return;
        }           
        
        try {
            // setup connection
            conn = doc.openConnection();
            conn.setUseCaches( false );
	    // Set the header so that the hyperlink URLs are not substituted by
	    //	the list of physical URLs and embedded resources
	    conn.setRequestProperty("DoNotSubstitute", "Yes");
            cont = conn.getContentType();
            in = conn.getInputStream();
        } catch( Exception e ){
            //System.out.println( "getDoc: in stream open " + e.getMessage() );
            return;
        }
	if (in == null) return;

	// Check if the target directory for storing the downloaded files exists.
	// If it does not exist, create the directory.
	File dirFd = new File( localDir );
	if( !dirFd.exists() ) {
	  try {
	    boolean dirCreated = dirFd.mkdir();
	    if (!dirCreated) {
	      System.out.println("Error: cannot create directory " + localDir);
	      return;
	    }
	  } catch (SecurityException se) {
	    System.out.println("Error: cannot create directory " + localDir);
	    return;
	  }
	}

        // extract documents from the network
        try {
	    extractDocs( in, localDir );
        } catch( Exception e ){
            //System.out.println( "getDoc: in extractDocs " + e.getMessage() );
            return;
        }
        
        try {
            in.close();
        } catch( Exception e ){
            //System.out.println( "getDoc: conn close " + e.getMessage() );
            return;
        }
    }

    // extractDocs() : extracts the retrieved resources and saves them
    //	on the local disk under localDir directory
    protected static void extractDocs( InputStream inStrm, String localDir )
        throws IOException
    {
        BufferedInputStream bInStrm; 
        byte buf[] = new byte[HUGE_BUF_LEN];
        byte newBuf[];
        byte tmpBuf[];
        BufferedOutputStream outStrm = null;
        int bytesRead;
        int totalRead = 0;
        String fileName = null;     
        int i;
        String base;
        Stack end = new Stack();        
       
        bInStrm = new BufferedInputStream( inStrm );
        while( (bytesRead = bInStrm.read( buf, totalRead, buf.length-totalRead)) 
               != -1 ){
            totalRead += bytesRead;
            
            if( totalRead >= buf.length ){
                // once starts filling up, increase size of buffer
                newBuf = new byte[buf.length + BUF_LEN];
                System.arraycopy( buf, 0, newBuf, 0, totalRead );
                buf = newBuf;
            }           
        }

	//System.out.println("Bytes read: " + totalRead);	// Mehmet: debug

        // extract and store files
        for( i = 0; i < totalRead; i++ ){
            if( i < buf.length - separatorB.length &&
                match( buf, i, separatorB ) ){

		String fname = getFileName( buf, i+separatorB.length, end );	// Mehmet: added this
                    
                fileName = localDir + fname;	// Mehmet: replaced getFileName() call with fname variable
                               
                if( outStrm != null ){
                    outStrm.close();
                }

                outStrm = new BufferedOutputStream( new FileOutputStream( 
                                                  fileName ));

/*
                if( fileName.indexOf( "html" ) != -1 ) {
                    base = "<BASE HREF=\"" + patchPath(localDir) + 
                           "\">\n";
                    tmpBuf  = base.getBytes();
                    outStrm.write( tmpBuf, 0, tmpBuf.length );
                }
*/
                
                // skip over the rest of separator
                i = ( (Integer) end.pop() ).intValue();

            } else {
	            // just copy bytes to a file if any open 
	            if( outStrm != null ){
                    outStrm.write( buf[i] );
                }
            }
        }

        if( outStrm != null ){
            outStrm.close();
        }

        return; 
    }

    protected static boolean match( byte buf[], int pos, byte sep[] )
    {
	int i;

	for( i = 0; i < sep.length; i++ ){
	  if( buf[pos+i] != sep[i] ){
	    return false;
          }
	}

	return true;
    }

    protected static String getFileName( byte buf[], int pos, Stack end )
    {
        int i;
        StringBuffer fileName = new StringBuffer();
        byte endString[] = new String( ">" ).getBytes();
        String tmpChar;

        Assert.assert( end.empty() );

        for( i = pos; i < buf.length; i++ ) {
	    if( match( buf, i, endString ) ) {
                end.push( new Integer(i) );
		String result = fileName.toString();
		int seppos = result.lastIndexOf('/');
		if (seppos != -1) return result.substring(seppos + 1);
		else return result;
            }

            tmpChar = new String( buf, i, 1 );
            fileName.append(tmpChar);
        }

        return null;
    }

/*    
    protected static String patchPath( String path )
    {
        path = "file:" + path;   
        
        if ( Webpp.os == Webpp.WIN ) {
            // HACK: we should write a generic string 
            // replacement routine to do this job
            // file:C:\Temp\file.html -> file:///C|/Temp/file.html
            path = path.replace( '\\', '/' );
            path = "file:///C|" + path.substring(7, path.length());
        }
        
        return path;
    }
*/

}
