package webppserv;

import java.net.*;
import java.io.*;
import java.util.*;
import javax.servlet.*;
import com.oroinc.text.regex.*;
import java.util.zip.*;		// Mehmet: added for checksum calculation
import java.lang.*;		// Mehmet: added for checksum calculation


///////////////////////////////////////////////////////////////////////////////
// 
// IO filter which does HTML HREF and SRC tag substitution.
//
///////////////////////////////////////////////////////////////////////////////
class SubstInputStream extends BufferedInputStream
{
   /*
    *=========================================================================
    * private class variables
    *=========================================================================
    */
    
    /* regex compiler machinery */ 
    static PatternMatcher matcher;
    static PatternCompiler compiler;
    protected static Pattern pattern[];
    protected static Pattern lineSepPattern;
    protected static Pattern bodyPattern;
    protected static Pattern srcPattern;
    
    /* substitution patterns */ 
    protected static String rule[];
    protected static String prefix;
    protected static String suffix;
       
    /* stream with parent doc */
    protected ByteArrayInputStream parentDocStream = null;

    /* cache file name */
    protected String cacheFile = null;
    protected BufferedOutputStream cacheFileStream = null;
    
    /* list of embedded resources */
    protected Hashtable embFileNames = new Hashtable();

    /* reference to a parent servlet */
    protected Webpp servlet;
    
    /* directory of parent html document */
    protected String parentDocDir;
    
    /* state of read() */
    private BufferedInputStream currentEmbStream = null;
    private Enumeration embDocEnum = null;
    private boolean parentDoc = false;

    /* Mehmet: Imported boolean from Webpp */
    protected boolean useCache = false;


   /*
    *=========================================================================
    * wrapper around the constructor which opens cache file out stream
    *=========================================================================
    */
    public SubstInputStream( String fileName, String cacheFileName, Webpp parentServlet ) 
	throws FileNotFoundException
    {
        this( fileName, parentServlet );
        
        Assert.assert( fileName != null && cacheFileName != null, 
                       fileName + " " + cacheFileName );

	if (parentServlet.useCache) {		// Mehmet: added
          cacheFile = cacheFileName;
          try {
            cacheFileStream = new BufferedOutputStream( 
                                           new FileOutputStream(cacheFile) );
          } catch( IOException e ) {
            Assert.assert( false, e.getMessage() );
          }
	  useCache = true;
	}
	else {
	  useCache = false;
	  cacheFile = null;
	  cacheFileStream = null;
	  System.out.println("Error: This fcn should be called only when Webpp.useCache is true!");
	}	// End Mehmet: added
    }


   /*
    *=========================================================================
    * object constructor
    *=========================================================================
    */
    protected SubstInputStream( String fileName, Webpp parentServlet )
        throws FileNotFoundException
    { 
        super( new FileInputStream(fileName) );
        
        String line = null;
        StringBuffer buf1 = new StringBuffer( Webpp.BUFFER_SIZE );
        String buf2 = null;
        //String res = null;
        StringBuffer res2 = new StringBuffer(204800);	// Mehmet: has to be big enough so that we don't allocate memory frequently
        String replace = null;
        String sep;
        String fileSep;
        String replList = null;
        PatternMatcherInput input = null;
        MatchResult result = null;
        int lastEnd = 0;
        String appletText = new String();
        String embFileName = null;
        BufferedReader parentDoc = null;
        String docType;
        boolean isHtml;
	long t0, t1 = 0, tt;		// Mehmet: measure times

        useCache = false;
        servlet = parentServlet;

        docType = servlet.getServletConfig().getServletContext().
                  getMimeType( fileName );
        isHtml = ( docType != null && docType.indexOf( "html" ) != -1 );
        
        sep = System.getProperty( "line.separator" );
        fileSep = System.getProperty( "file.separator" );
 
	try {
	  FileReader dummyFR = new FileReader( fileName );
	  if (dummyFR != null)
            parentDoc = new BufferedReader(dummyFR);
	  else parentDoc = null;
	} catch (Exception er) {
	  System.out.println("Error: cannot open file: " + fileName);
	  parentDoc = null;
	}
        parentDocDir = fileName.substring( 0, 
                                           fileName.lastIndexOf(fileSep)+1 );

	if (parentDoc != null) {
          try {
            while( (line = parentDoc.readLine()) != null ){
                if( buf1 == null ){
                    buf1.append( line );
                } else {
                    buf1.append( sep );
                    buf1.append( line );
                }
            }
          } catch( IOException e ){
            Assert.assert( false, e.getMessage() + "when reading " + 
                                  fileName );
          }
	  catch(Exception e2) {
	    System.out.println("Error: exception when reading " + fileName);
	  }
	}
        
        //netChecksumTime = 0;	// Mehmet: Measure times

        // add reference to an applet just after <BODY> 
        appletText = 
            "<BODY> \n" +
            "<SCRIPT LANGUAGE=\"JavaScript\"> \n" +
            "document.writeln( '<applet name=Webpp' ); \n" +
            "document.writeln( 'code=Webpp.class archive=webpp.jar' ); \n" +
            "document.writeln( 'codebase=http://" +
            Webpp.server + ":" + Webpp.port + "' ); \n" +
            "document.writeln( 'width=2 height=2>' ); \n" +
            "document.writeln( '<param name=\"cabbase\" value=\"webpp.cab\">' ); \n" +
            "document.write( '<param name=\"env\" value=\"' ); \n" +
            "document.writeln( navigator.appName + ' ' + navigator.appVersion + '\">' ); \n" +
            "document.writeln( '<param name=\"trace\" value=\"true\">' ); \n" +
            "document.writeln( '<param name=\"timing\" value=\"true\">' ); \n" +
            "document.writeln( '</applet>' ); \n" +
            "</SCRIPT> \n";
        
        if( isHtml ){
            buf2 = Util.substitute( matcher, bodyPattern, appletText, 
                                    buf1.toString() );
        } 
        
        // parse the rest of document for HREF
        for( int i = 0; isHtml && i < pattern.length; i++ ){
            input = new PatternMatcherInput( buf2 );

            //res = null;
            res2.setLength(0);		// clear res2 contents
            lastEnd = 0;
            while( matcher.contains(input, pattern[i]) ){
                result = matcher.getMatch(); 
                
                if( result.group(1).indexOf( "javascript:") != -1 ||
                    result.group(1).indexOf( ".css" ) != -1 ){
                    // substitution in javascript code does not make
                    // sense and leads to errors in HTML
                    // similarly, no substitution should be done for CSS
                    continue;
                }
                                
                /************************************************************** 
                   check replication directory for matches                   
                   FIXME: when we use prefix addressing, needs to be replaced 
                          with iteration over all directory prefixes starting
                          with the maximal one
                **************************************************************/
                replList = ReplicationDir.getUrls( result.group(1) );
                
                // if no match found, leave the HREF alone
                if( replList == null ){
                    continue;
                }

		// Begin Mehmet
		// Mehmet: added the following part to substitute the emb resource names with
		// 	checksums. If Webpp.useChecksum is true, this operation will be
		// 	carried on. Otherwise, client will receive the real names of resources.
		t0 = System.currentTimeMillis();	// Mehmet: Measure times
		if (parentServlet.useChecksum ) {
		  int pos = replList.indexOf(' ');
		  String embResStr = replList.substring(0, pos);
		  String replList2 = replList.substring(pos+1);
		  String nextDocDir = null;	// directory of the pointed URL on this server, if exists
		  // If the pointed URL is also replicated on the current server, get the
		  // directory path of that URL. This is used for checksum calculation.
		  if (replList2 != null) {
		    StringTokenizer st2 = new StringTokenizer(replList2);
		    for ( ; st2.hasMoreElements(); ) {
		      String serverStr = (String) st2.nextElement();
		      //System.out.println("serverStr=" + serverStr);	// Mehmet: debug
		      //System.out.println(serverStr.substring(0,25));
		      //System.out.println(serverStr.substring(25));
		      //System.out.println("parentServlet.server=" + parentServlet.server);
		      int strPos = serverStr.indexOf(parentServlet.server);
		      //System.out.println("strPos=" + strPos);
		      if (strPos != -1) {	// this server has a replica of the pointed URL
			int dirPos = serverStr.toLowerCase().indexOf("webppserv.webpp");	// all lower case just to be sure
			//System.out.println("dirPos=" + dirPos);
			if (dirPos != -1) {
			  dirPos += 16;		// bypass "webppserv.Webpp" part of the URL
			  int endPos = serverStr.lastIndexOf("/");
			  if (endPos == -1) endPos = serverStr.lastIndexOf(System.getProperty("file.separator"));
			  //System.out.println("endPos=" + endPos);
			  if (endPos == -1 || endPos <= dirPos)
			    nextDocDir = parentServlet.rootDir;
			  else
			    nextDocDir = parentServlet.rootDir + serverStr.substring(dirPos, endPos) +
						System.getProperty("file.separator");
			}     // end if (dirPos != -1)
		      }	    // end if (strPos != -1)
		    }	  // end for ( ; st2.hasMoreElements(); )
		  }	// end if (replList2 != null)
		  tt = System.currentTimeMillis() - t0;	// Mehmet: Measure times
		  //System.out.println("Time1: " + tt);		// Mehmet: Measure times
		  t1 += tt;					// Mehmet: Measure times

		  //System.out.println("nextDocDir=" + nextDocDir);	// Mehmet: debug
		  //System.out.println(nextDocDir.substring(0,20));
		  //System.out.println(nextDocDir.substring(20));

		  String newList = new String();
		  if (nextDocDir != null && embResStr != null && !embResStr.equalsIgnoreCase("NULL"))
		  {
		    StringTokenizer st = new StringTokenizer(embResStr, ";");
		    for ( ; st.hasMoreElements(); ) {
		      String resource = (String) st.nextElement();
		      t0 = System.currentTimeMillis();		// Mehmet: Measure times
		      String chksum = parentServlet.getChecksumValue(new Adler32(), nextDocDir + resource);
		      tt = System.currentTimeMillis() - t0;	// Mehmet: Measure times
		      //System.out.println("Time2: " + tt);	// Mehmet: Measure times
		      t1 += tt;
		      if (chksum == null) chksum = resource;
		      if (newList.length() > 0) newList += ";" + chksum;
		      else newList += chksum;
		    }
		    replList = newList + " " + replList2;
		  }
		}	// end if (parentServlet.useChecksum)
		// End Mehmet
                
                res2.append(buf2.substring( lastEnd, result.beginOffset(0) ) );
                res2.append(rule[2*i+1] + prefix + replList + suffix);
                
                lastEnd = result.endOffset(0);
            }	// while(matcher.contains(input, pattern[i]))
            
            if( res2.length() > 0 ){
                // append the rest of string 
                res2.append( buf2.substring( lastEnd, buf2.length() ) );
                // use new line in next matching 
                buf2 = res2.toString();
            }
        }
                   
        // parse all SRC and generate a vector with file names 
        if( isHtml ){
            input = new PatternMatcherInput( buf2 );
	    //res = null;		// Mehmet: added for substituting emb res names with checksums
	    res2.setLength(0);		// clear res2 buffer contents
	    lastEnd = 0;	// Mehmet: for checksum substitution
            while( matcher.contains(input, srcPattern) ){
                result = matcher.getMatch();
                embFileName = parentDocDir + result.group(1);

		String fname = new String();	// Mehmet: for storing the checksum

		// Mehmet: added the following if stmt to calculate checksum if necessary
		if (parentServlet.useChecksum ) {
		  t0 = System.currentTimeMillis();		// Mehmet: Measure times
		  fname = parentServlet.getChecksumValue(new Adler32(), embFileName);
		  tt = System.currentTimeMillis() - t0;		// Mehmet: Measure times
		  //System.out.println("Time3: " + tt);		// Mehmet: Measure times
		  t1 += tt;
		}
		else
		  fname = result.group(1);
		// End Mehmet
                
                // the use of hashtable is important since we
                // do not want to send duplicates 

		// ***** Begin Mehmet *****
		// Mehmet: I added the following if statement to exclude the embedded
		// resources that are already in the client's cache.
		if ((parentServlet.clientHas == null) || (parentServlet.clientHas.indexOf(";" + fname + ";")) == -1) {
		  //System.out.println("Mehmet: Packaging: " + result.group(1) + " (" + fname + ")");	// Mehmet: debugging
		  embFileNames.put( embFileName, new String( " " ) );
		}
		//else {
		//  System.out.println("Mehmet: Excluding: " + result.group(1) + " (" + fname + ")");	// Mehmet: debugging
		//}

		// Mehmet: I added the following if stmt to substitute embedded resource names
		// 	with checksums when useChecksum flag is set to true in Webpp class.
		if (parentServlet.useChecksum ) {
		  t0 = System.currentTimeMillis();		// Mehmet: Measure times
                  res2.append(buf2.substring( lastEnd, result.beginOffset(0) ) );
		  res2.append("SRC = \"" + fname + "\"");
                  lastEnd = result.endOffset(0);
                  tt = System.currentTimeMillis() - t0;		// Mehmet: Measure times
                  //System.out.println("Time4: " + tt);		// Mehmet: Measure times
                  t1 += tt;
		}	// end if (parentServlet.useChecksum)
		// ***** End Mehmet *****
	    }

	    // begin Mehmet: added the following also for checksum substitution
	    if (res2.length() > 0) {
	      t0 = System.currentTimeMillis();		// Mehmet: Measure times
              // append the rest of string 
              res2.append( buf2.substring( lastEnd, buf2.length() ) );
              // use new line in next matching 
              buf2 = res2.toString();
              tt = System.currentTimeMillis() - t0;	// Mehmet: Measure times
              //System.out.println("Time5: " + tt);	// Mehmet: Measure times
              t1 += tt;
            }
	    // end Mehmet

            embDocEnum = embFileNames.keys();
        }
        
        if( isHtml && parentDoc != null ){
            parentDocStream = new ByteArrayInputStream( buf2.getBytes() ); 
        } else if (parentDoc != null) {
            parentDocStream = new ByteArrayInputStream( 
                                              buf1.toString().getBytes() );
        }
	else parentDocStream = null;

	if (parentDoc != null) {
          try {
            parentDoc.close();
          } catch( IOException e ){
            Assert.assert( false, "closing " + fileName );
          }
        }
        
	if (parentServlet.traceSuboperationTimes) {
	  System.out.println("Net Checksum Time: " + parentServlet.netChecksumTime);	// Mehmet: Measure times
	  //System.out.println("Total Checksum Time: " + t1);		// Mehmet: Measure times
	}
    }


   /*
    *=========================================================================
    * overloaded - iterates over the parent doc file and files of all emb docs
    *=========================================================================
    */
    public int read( byte buf[], int pos, int len) throws IOException
    {  
        int bytesRead;
        String embDocFileName;
        String separator;
        String docName;
        byte tmpBuf[];
        
	if (parentDocStream == null) return 0;

        /* first send embedded resources */
        if( currentEmbStream != null &&
            (bytesRead = currentEmbStream.read( buf, pos, len )) != -1 ){
	    if (useCache)		// Mehmet: added
              cacheFileStream.write( buf, pos, bytesRead );
            
            return bytesRead;
        }
         
        /* open another emb resource file */
        while( embDocEnum != null && embDocEnum.hasMoreElements() ){
            embDocFileName = (String) embDocEnum.nextElement();
            
            if( currentEmbStream != null ){
                currentEmbStream.close();
            }
            try {
                currentEmbStream = new BufferedInputStream( 
                                      new FileInputStream( embDocFileName ) );
            } catch( FileNotFoundException e ){
                System.out.println( "emb. resource " + embDocFileName +
                                    " not found" );
                continue;
            }
            
            /* prepare a separator to be put between the files */
	    if (Webpp.useChecksum ) {
	      docName = Webpp.getChecksumValue(new Adler32(), embDocFileName);
	    }
	    else {
	      docName = embDocFileName.substring( parentDocDir.length(), 
						embDocFileName.length() );
	    }

	    /* Put the separator between the files */
            separator = "<WEBPP_APPEND " + docName + ">";
            tmpBuf = separator.getBytes();
            Assert.assert( tmpBuf.length > 0 );
            System.arraycopy( tmpBuf, 0, buf, 0, tmpBuf.length );
	    if (useCache)		// Mehmet: added
              cacheFileStream.write( tmpBuf, 0, tmpBuf.length );
            
            return tmpBuf.length;
        }

	// Begin Mehmet: if client has the parent document in its cache, don't send it to client
	if (servlet.clientHasParentDoc) return -1;
	// End Mehmet
        
        /* add separator for parent document */
        if( !parentDoc ){
            Assert.assert( servlet.reqFile != null );
            docName = servlet.reqFile.substring( Webpp.rootDir.length(),
                                                 servlet.reqFile.length() );
            separator = "<WEBPP_APPEND " + docName + ">";
            tmpBuf = separator.getBytes();
            Assert.assert( tmpBuf.length > 0 );
            System.arraycopy( tmpBuf, 0, buf, 0, tmpBuf.length );
            
	    if (useCache)		// Mehmet: added
              cacheFileStream.write( tmpBuf, 0, tmpBuf.length );            
            parentDoc = true;

            return tmpBuf.length;
        }
        
        /* read while you can from parent document */
        bytesRead = parentDocStream.read( buf, pos, len );
        if( bytesRead != -1 ){
	    if (useCache)		// Mehmet: added
              cacheFileStream.write( buf, pos, bytesRead );
            
            return bytesRead;
        }
        
        /* everything has been read */
        if( currentEmbStream != null ){
            currentEmbStream.close();
        }
	if (useCache)		// Mehmet: added
          cacheFileStream.close();

        return -1;
    }
                   

   /*
    *=========================================================================
    * class constructor
    *=========================================================================
    */
    static 
    {
        int i = -1;

        /* initialize regex compiler */
        matcher = new Perl5Matcher();
        compiler = new Perl5Compiler();

        /* initialize substitution rules */
        rule = new String[] { "HREF\\s*=\\s*\"(\\S+)\"",
                              "HREF=\"\" onClick=\"this.href = " };
        prefix = "document.Webpp.getUrl(\'";
        suffix = "\')\"";

        pattern = new Pattern[rule.length/2];

        /* parse regular expressions */
        try {
            for( i = 0; i < rule.length; i += 2 ){
                pattern[i/2] = compiler.compile( rule[i],
                                       Perl5Compiler.CASE_INSENSITIVE_MASK );  
            }
            lineSepPattern = compiler.compile( "\n+" );
            bodyPattern    = compiler.compile( "<.*BODY.*>", 
                                       Perl5Compiler.CASE_INSENSITIVE_MASK );
            srcPattern = compiler.compile( "SRC\\s*=\\s*\"(\\S+)\"",
                                       Perl5Compiler.CASE_INSENSITIVE_MASK );
                                       
        } catch( Exception e ){
            Assert.assert( false, Integer.toString(i) + e.getMessage() );
        }
    }
}
