package webppserv;
// Web++ Server

// The initialization is done when one of the following happens:
//  - A get request is received and the initialization has not been
//      done before
//  - A post request is received and it sets the TestPhase to zero
//      (note that in this case, the initialization is done 
//	even if those were already done before.
//      Thus, a TestPhase call with phase number 0 is used for
//      restarting the tests.

import java.net.*;
import java.io.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import com.oroinc.text.regex.*;
import java.util.zip.*;		// Mehmet: added for checksum calculation
import java.lang.*;		// Mehmet: added for checksum calculation



///////////////////////////////////////////////////////////////////////////////
//
// Servlet implementing the Webpp functionality
//
///////////////////////////////////////////////////////////////////////////////
public class Webpp extends HttpServlet 
{
    /* class variables */
    static final int BROWSERTEST = 1;   // this test is implemented for running on browsers
    static final int AUTOMATICTEST = 2; // this test is implemented for running from automatic scripts
    static final int TestType = AUTOMATICTEST;
    static final String NAME_SERVER = "begonia.ece.northwestern.edu:8080";	// main server (name server, a.k.a. coordinator)
    final static String NAME_SERVER_IP = "129.105.5.141:8080";	// IP address of the name server
    static final int BUFFER_SIZE = 8192;   // buffer size used to read files
    // min no of requests needed for a resource before starting to make replication decisions
    static long MinReqCountForDecision = 10;
    // min load percentage to make a replication decision
    static double MinLoadPercentageForDecision = 0.005;
    // min load percentage out of totalRequestCount to consider a document as hot
    static double MinReqPercentageForHotDoc = 0.002;
    // a clear for all counts is enforced when totalRequestCount reached the following number
    static long LimitToEnforceClearCount = 2000;

    static final int TRACEROUTE = 1;	// constant for this segmentation method
    static final int LATENCY = 2;		// constant for this segmentation method

    static long testStartTime = 0;	// start time of the test in local time
    static String aliveServers = null;	// keeps alive servers info
    static String logStr;       // string to be written to log
    static String logfname = "public_html/logfile.txt"; // filename for the log file
    //static PrintWriter logfile = null;
    static RandomAccessFile logfile = null;
    static int testPhase = 0;
    static long overallTime = 0;	// total time spent for serving a request
    static long totalRequestCount = 0;	// total no of requests received for all resources
    static String rootDir = null;   // document root directory of the server
    static String webppRootDir;                      // webpp root directory
    static String server = null;                              // server name
    static int serverID = -1;			// Mehmet: server ID number
    static int port = -1;                              // server port number
    // Mehmet: clientHas string lists embedded resources of the requested URL
    // that are in client cache.
    String clientHas = null;
    
    static int tracerouteCount;	// keep track of # of threads for traceroute
    static int msgCount;	// keep track of # of threads for msg broadcasting
    static int segmentationMethod = LATENCY;	// how server segments are built (using TRACEROUTE or LATENCY)

    static boolean useCache = false;                    // if set use cache
    static boolean debugOutput = false;        // print output stream if set
    static boolean debugReplDir = false;     // debug msgs for repl dir ops
    static boolean debugCache = false;   // debug msgs related to cache use
    static boolean useChecksum = true;  // Mehmet: true -> replace emb res names with checksums
    static boolean clientHasParentDoc = false;	// Mehmet: will be set using request headers. If true, don't send parent doc.
    static boolean traceSuboperationTimes = false;	// Mehmet: measure times
    static boolean serversInitialized = false;	// to make sure we initialize other servers only once
    static boolean traceroutesDone;	// to make sure servlet creates its tree structure view and segments
    static boolean traceroutesStarted;	// to make sure we don't try to start traceroutes when they are in progress but but done yet
    static boolean tracerouteDisabled = false;	// set if a problem occurs with traceroute command so that the rest of programs do not use it
    static boolean latenciesStarted;	// to make sure we don't try to start latency measurement when it is in progress but but done yet
    static boolean latenciesDone;	// to make sure servlet measures its distance to all other servers
    static int requestCount = 0;	// count # of ClientRequest requests processed

    // doNotSubstituteHyperlinks is normally set to false. It is set to true only
    //	when the "DoNotSubstitute" header is received from the client, which means that
    //	this is not an ordinary client request but it is a request from another server
    //	by which the resources are downloaded onto that server (new replica or update).
    static boolean doNotSubstituteHyperlinks = false;

    /* Mehmet: Measure net checksum time for reading files and calculating
	checksums */
    static long netChecksumTime = 0;
    static Integer clientSegment = null;	// segment where the current request's initiator client belongs
    /* Mehmet: use a hash table to store calculated checksums during the
	service of the current request in order to avoid multiple checksum
	calculations for an embedded resource file */
    static ChecksumTable checksumHash = null;
    static Hashtable serverIDs = null;		// ID numbers of all servers
    static Hashtable serverPorts = null;	// port number used by each Web++ server
    static Hashtable serverStatus = null;	// is set to "Alive" or "Dead" for each server
    static Hashtable clientIDs = null;         // ID numbers of all clients
    static Hashtable clientPorts = null;       // port number used by each Web++ client 
    static Hashtable clientStatus = null;      // is set to "Alive" or "Dead" for each client
    static Hashtable traceroutePaths = null;	// stores the traceroute paths to every other Web++ serevr
    static Hashtable HostServer = null;	// keeps track of which server is host for each client
    static SerHashtable assigned = null;	// tells which servers are assigned to which segments (Serializable Hash Table)
    static Hashtable segmentSizes = null;	// contains sizes of segments
    static ResponseTimeTable respTimes = null;	// response time info for each document and segment (server group)

    static int noOfSegments = 0;	// no of segments (server groups)
    static boolean BrowserNeedsData = false;	// Is set to true if servlet recognizes that browser cannot handle an empty response
    static String prevResponseTime = null;	// stores previous resp time piggybacked that is by client
    
    /* package variables */
    String reqFile = null; 
  
    /* private variables */
    protected static Cache cache = null;                     // cache object
    protected static ReplicationDir repl = null;   // replication dir object
    protected static Pattern parmSepPattern = null;  // param separator patt
    

   /*
    *=========================================================================
    * Constructor
    *=========================================================================
    */    
    public void init( ServletConfig config ) throws ServletException
    {
        super.init( config );
    }	// init

    public void initWebppServer(String caller) throws ServletException
    {
	traceroutesDone = false;	// Mehmet: to solve problem with Germany
	traceroutesStarted = false;	// Mehmet: to solve problem with Germany
	latenciesStarted = false;	// Mehmet: to solve problem with Germany
	latenciesDone = false;	// Mehmet: to solve problem with Germany

	System.out.println("In initWebppServer");

	if (NAME_SERVER.equals(server + ":" + port) ||
                NAME_SERVER_IP.equals(server + ":" + port))
        {
          //System.out.println("This is NAME SERVER");
          serverID = 0;
	  testStartTime = System.currentTimeMillis();
	}
	else {
	  testStartTime = getTestTimeFromNameServer();
          if (testStartTime == 0) testStartTime = System.currentTimeMillis();
	  serverID = -1;
	}

        webppRootDir = System.getProperty( "server.root" ) + 
                       System.getProperty("file.separator" ) + 
                       "webpp";
        
	if (cache == null) cache = Cache.factory( this );
	else cache.clear();
	if (repl == null) repl = ReplicationDir.factory();
	//else repl.clear();
	if (checksumHash == null) checksumHash = ChecksumTable.factory();
	else checksumHash.clear();
	if (assigned == null) assigned = SerHashtable.factory("segments.ser", this);
	//else assigned.clear();

	if (serverIDs == null) serverIDs = new Hashtable();
	else serverIDs.clear();
	if (serverPorts == null) serverPorts = new Hashtable();
	else serverPorts.clear();
	if (serverStatus == null) serverStatus = new Hashtable();
	else serverStatus.clear();
	if (clientIDs == null) clientIDs = new Hashtable();
        else clientIDs.clear();
        if (clientPorts == null) clientPorts = new Hashtable();
        else clientPorts.clear();
        if (clientStatus == null) clientStatus = new Hashtable();
        else clientStatus.clear();
	if (traceroutePaths == null) traceroutePaths = new Hashtable();
	else traceroutePaths.clear();
	if (HostServer == null) HostServer = new Hashtable();
	else HostServer.clear();
	if (segmentSizes == null) segmentSizes = new Hashtable();
	else segmentSizes.clear();
        if (assigned != null && !assigned.isNew()) {
          int segCount = getNoOfSegmentsFromTable();
          noOfSegments = segCount;
        }
	if (respTimes == null) respTimes = ResponseTimeTable.factory(noOfSegments);
	else respTimes.clear();
	
        try {
            parmSepPattern = SubstInputStream.compiler.compile( "&" );
        } catch( MalformedPatternException e ){
            throw new ServletException( "malformed parameter separator" );
        }

	// Create or open the log file
	File fd = new File( logfname );
	if( fd.exists() ) {
	  // Open the existing log file for append
	  try {
	    logfile = new RandomAccessFile(fd, "rw");
	  } catch (IOException e) {
	    System.out.println("Error: cannot open logfile for read-write");
	  }
	  try {
	    logfile.seek(logfile.length());	// go to end of file (Append mode)
	  } catch (IOException e) {
	    System.out.println("Error: cannot seek end of log file");
	  }
	}
	else {
	  // Create a new log file
          try {
            //logfile = new PrintWriter(new BufferedWriter(new FileWriter(logfname)));
	    logfile = new RandomAccessFile(logfname, "rw");
          } catch( IOException e ){
            System.out.println( "Error: cannot create log file: " + e.getMessage() );
          }
	}
	
	readOptionsFile();
        testPhase = 0;
        writeToLogWithTime("Start");
	if (serverID == 0)
        {
          readServerList();
          readClientList();
          initServers(caller);
          if (segmentationMethod == LATENCY && !latenciesStarted && !latenciesDone)
            doLatencies(this);
        }

    }	// initWebppServer()


   /*
    *=========================================================================
    * Write the given string to the log file
    *=========================================================================
    */
    static synchronized void writeToLog(String str)
    {
        if (str == null) {
          System.out.println("Error: str is null in writeToLog");
          return;
        }

        try {
          logfile.writeBytes(str);
        } catch (IOException e) {
          System.out.println("Error: cannot write to logfile: " + e);
        }
    }

   /*
    *=========================================================================
    * Write the given string to the log file after adding standard log
    *  info at the beginning of it (e.g. current test time and server name)
    *=========================================================================
    */
    static void writeToLogWithTime(String str)
    {
      if (str == null) {
        System.out.println("Error: str is null in writeToLog");
        return;
      }

      str = Long.toString(System.currentTimeMillis() - testStartTime) + " Server " + server + " " + str + "\n";
      writeToLog(str);
    }

   /*
    *=========================================================================
    * Destroy the current Web++ server
    *=========================================================================
    */
    static public void destroyServer() {
        System.out.println("Destroying Server");
	writeToLogWithTime("Destroy");
        System.exit(1);
    }
         
   /*
    *=========================================================================
    * read documents with a substitution.
    * syntax: GET http://serverIP/servlet/webppserv.Webpp/<filename>
    *         (<filename> is relative to servlet home dir)
    *=========================================================================
    */    
    protected void doGet( HttpServletRequest req,
                          HttpServletResponse resp ) 
        throws ServletException, IOException
    {
        BufferedReader reader = null;
        ServletOutputStream out = null;
        BufferedInputStream in = null;
        String mimeType = null;
        byte buf[] = new byte[BUFFER_SIZE];
        int bytesRead;
        Date now;
        long t0 = 0, t1 = 0;	// Mehmet: Measure times

	//System.out.println("ReqType: " + req.getMethod());

	// Begin Mehmet: check query string first
	clientHas = req.getQueryString();
	String mainReq = req.getPathInfo();
	System.out.println("\nDocument: " + mainReq);
	System.out.println("Server name: " + server);
	//System.out.println("clientHas: " + clientHas);
	if (clientHas != null && ((clientHas.indexOf("ServerList") >=0) ||
		(clientHas.indexOf("HostServer") >=0) ||
		(clientHas.indexOf("Locations") >=0) ||
		(clientHas.indexOf("TestPhase") >=0) ||
		(clientHas.indexOf("AddToReplDir") >=0) ||
		(clientHas.indexOf("HostServerBroadcast") >=0) ||
		(clientHas.indexOf("AliveServers") >=0) ||
		(clientHas.indexOf("SendAliveServers") >=0) ||
		(clientHas.indexOf("GetTestTime") >=0) ||
		(clientHas.indexOf("DownloadReplicationDir") >=0) ||
		(clientHas.indexOf("GetReplicationDir") >=0) ||
		(clientHas.indexOf("ReplicationRequest") >=0) ||
		(clientHas.indexOf("Destroy") >=0)) ) {
	  doPost(req, resp);
	  return;
	}

        if( port == -1 ){
            port = req.getServerPort();
        }
        if( rootDir == null ){
            rootDir = req.getRealPath( "/" );
        }
	if( server == null ){
            server = req.getServerName();
            initWebppServer("doGet");          // initialize server
        }
        //System.out.println("Server " + server + " is up");

	// Mark the beginning of overallTime measurement
	overallTime = System.currentTimeMillis();

	// If there is info about previous request's resp time, extract it
	prevResponseTime = null;
	if (clientHas != null) {
	  int pos1 = clientHas.indexOf('|');
	  if (pos1 >= 0) {
	    int pos2 = clientHas.lastIndexOf('|');
	    if (pos2 > pos1) {
	      prevResponseTime = clientHas.substring(pos1+1, pos2);
	      System.out.println("prevResponseTime: " + prevResponseTime);
	    }
	    else {
	      System.out.println("Error: cannot extract prev response time properly!");
	    }
	    if (pos1 == 0) clientHas = null;
	    else clientHas = clientHas.substring(0, pos1);
	    //System.out.println("New clientHas: " + clientHas);
	  }
	}
	// End Mehmet

	// Begin Mehmet: find the client's segment if you can.
	String clientAddr = req.getRemoteAddr();
	String clientName = req.getRemoteHost();
	clientSegment = getSegmentofClient(clientAddr, clientName);
	if (clientSegment != null)
	  System.out.println("Segment of client " + clientAddr + "(" + clientName + ") is " + clientSegment);
	// End Mehmet

	// Begin Mehmet: if prevRespTime and clientSegment could be determined,
	//	then insert the information into the respTimes hash table.
	if (clientSegment != null && prevResponseTime != null)
	  insertRespTime(clientSegment, prevResponseTime);
	// End Mehmet

	// Begin Mehmet: debug headers
	//Enumeration reqHeaders = req.getHeaderNames();
	//while (reqHeaders != null && reqHeaders.hasMoreElements()) {
	//  String headerStr = ((String) reqHeaders.nextElement());
	//  String valueStr = null;
	//  if (headerStr != null) 
	//    valueStr = req.getHeader(headerStr);
	//  if (headerStr != null && valueStr != null)
	//    System.out.println("Header: " + headerStr + " : " + valueStr);
	//}
	// End Mehmet: debug headers

	t0 = System.currentTimeMillis();	// Mehmet: Measure times

	netChecksumTime = 0;	// Mehmet: Measure times

	// If this is this the main server, read data from disc
	if (NAME_SERVER.equals(server + ":" + port) ||
		NAME_SERVER_IP.equals(server + ":" + port))
	{
	  //System.out.println("This is NAME SERVER");
	  serverID = 0;
	  //System.out.println("Server-" + serverID + " at " + server + ":" + port + " is up");
	}
	else if (testStartTime == 0) {
	  testStartTime = getTestTimeFromNameServer();
	  if (testStartTime == 0) testStartTime = System.currentTimeMillis();
	  readOptionsFile();
	  testPhase = 0;
	  //System.out.println("Server is " + server + ":" + port);
	  serverID = -1;
	  writeToLogWithTime("Start");
	}

	// Mehmet: initialize all servers so that they create server clusters
	if (serverID == 0 && !serversInitialized) {
	  testStartTime = System.currentTimeMillis();
	  readServerList();
	  readClientList();
	  testPhase = 0;
	  readOptionsFile();
	  initServers("DoGet");
	  // Mehmet: I commented out the following since it puts too much
	  //  burden on the NAME_SERVER at the beginning of the test.
	  //  Instead, I wrote a shell script called 'dowakeupclients'
	  //  which wakes up all the clients. You chould call that script
	  //  after NAME_SERVER wakes up all other Web++ servers. 
	  //initClients(this);        // also initialize client servlets
	  if (segmentationMethod == LATENCY && !latenciesStarted && !latenciesDone)
	    doLatencies(this);
	}

	// Mehmet: do traceroutes to create tree structure view of network
        if (segmentationMethod == TRACEROUTE && !traceroutesStarted && !traceroutesDone) {
	  doTraceroutes(this);
        }

        logStr = Long.toString(System.currentTimeMillis() - testStartTime);
	if (clientName == null) clientName = "null";
        logStr += " Server " + server + " ClientRequest";
	          
        reqFile = req.getPathTranslated();
        if( reqFile == null ){
  	  System.out.println("Error: reqPath cannot be translated!");
          throw new ServletException( "no extra path or could not xlate" );
        }

        // Mehmet: added the following if statement to avoid assert
        //  crashes (April 24 2001)
        if (reqFile.endsWith("/"))
          reqFile += "index.html";

	String tmpstr = null;
	if (mainReq != null && mainReq.startsWith("/") && mainReq.length() > 1)
	  tmpstr = mainReq.substring(1);	// skip the first '/'
	else tmpstr = mainReq;
        logStr += " " + tmpstr + " " + clientAddr + " " + clientName;
        if (clientSegment != null) {
    	  logStr += " " + clientSegment.toString();
        }
        logStr += " " + Integer.toString(serverID);

        // Begin Mehmet
	//System.out.println("Mehmet: reqFile:" + reqFile);	// Mehmet: debug
	//System.out.println("Document: " + mainReq);		// Mehmet: debug
	//System.out.println("Client has: " + clientHas);	// Mehmet: debug
	String tmp = req.getHeader("DoNotSubstitute");
	if (tmp != null && tmp.equalsIgnoreCase("Yes"))
	  doNotSubstituteHyperlinks = true;
	else doNotSubstituteHyperlinks = false;
	tmp = req.getHeader("ClientHasParentDoc");
	if (tmp == null) {	// The browser might be Netscape, so check clientHas
	  //System.out.println("Header NOT received");
	  if (clientHas != null && clientHas.indexOf(";parent;") != -1) {
	    BrowserNeedsData = true;
	    tmp = "yes";
	  }
	}
	else {
	  //System.out.println("Header received");
	  BrowserNeedsData = false;	// Browser is not Netscape, so it can handle empty response properly
	}
	if (tmp != null && tmp.equalsIgnoreCase("yes")) {
	  //System.out.println("Client has parent doc");	// Mehmet: debug
	  clientHasParentDoc = true;
	}
	else {
	  //System.out.println("Client does not have parent doc");	// Mehmet: debug
	  clientHasParentDoc = false;
	}
	// End Mehmet

        mimeType = getMimeType( reqFile );
        resp.setContentType( mimeType );
	//resp.setHeader("TestHeader","nanik");		// Mehmet: debug headers

	  // Begin Mehmet: replaced the following if-else block with the
	  //	resp.setDateHeader call below. I do not want the indexWebpp.html
	  //	to be cached by the client any more.
	  resp.setDateHeader( "Expires", 0 );
/*
        if( reqFile.indexOf("indexWebpp.html") != -1 ){
            // set Expires for root document to 30 min from now
            //     allows browsers to return to parent docs even when 
            //     the server is down
            now = new Date();
            resp.setDateHeader( "Expires", now.getTime() + 1800000 );
	    // Begin Mehmet: get the list of server names and send them to the
	    //	client.
	    //String serverNamesStr = getActiveServerNamesList();
	    //if (serverNamesStr != null)
	    //  resp.setHeader("ServerList", serverNamesStr);	// Mehmet send the list of server names with port numbers
	    // End Mehmet
        } else {
            // all other docs are non-cacheable
            resp.setDateHeader( "Expires", 0 );
        }
*/
        // End Mehmet

        out = resp.getOutputStream();

	// Begin Mehmet: check if client has all emb resources and parent doc.
	// If so, then skip the rest of the operations so that client can
	// immediately its cached copies of resources.
	if (clientHasParentDoc && clientHas.equalsIgnoreCase("ALL")) {
	    //System.out.println("Client has ALL, so servlet does not send anything");	// Mehmet: debug
	    //writeToLog(logStr + "\n");
	    return;
	}
	// End Mehmet

	// Begin Mehmet: if client cache contains the parent document, check if it also
	//	contains all embedded resources. If it does, then simply skip the rest
	//	of the procedures and return nothing to the client. Client will use the
	//	local copies automatically.
	String fileSep = System.getProperty( "file.separator" );
	String parentDocDir = reqFile.substring( 0, 
        			reqFile.lastIndexOf(fileSep)+1 );

	if (clientHasParentDoc) {
	  String EmbResList = ReplicationDir.getEmbeddedResources( reqFile );
	  boolean hasAll = true;
	  if (EmbResList != null) {
	    StringTokenizer st = new StringTokenizer(EmbResList, ";");
	    for ( ; hasAll && st.hasMoreElements(); ) {
	      if (useChecksum)
		tmp = getChecksumValue(new Adler32(), parentDocDir +
					((String) st.nextElement()));
	      else
		tmp = ";" + ((String) st.nextElement()) + ";";
	      if (clientHas.indexOf(tmp) == -1) hasAll = false;
	    }
	  }
	  if (hasAll) {
	    //System.out.println("Client has all resources and parent doc");
	    if (BrowserNeedsData) {	// Browser cannot handle an empty response (most probably Netscape)
	      out.print("UseLocalCopy");	// Don't send empty response, just send something
	      out.close();
	    }
            //writeToLog(logStr + "\n");
	    return;
	  }
	}
	// End Mehmet

	t1 = System.currentTimeMillis() - t0; // Mehmet: time for processing the request headers
	if (traceSuboperationTimes) {
	  //System.out.println("\nProcessing Headers: " + t1);	// Mehmet: Measure times
	}
	logStr += " " + Long.toString(t1);
	t0 = System.currentTimeMillis();      // Mehmet: Measure time for next suboperation (preparing the batch file)

        try {
	    if (doNotSubstituteHyperlinks)
	      in = new BatchInputStream(reqFile, this);
	    else {
	      if (useCache)
                in = cache.getFile( reqFile );	// Mehmet: inserted if-else to this line
	      else
		in = new SubstInputStream( reqFile, this );
	    }
        } catch( FileNotFoundException e ){
            resp.sendError( HttpServletResponse.SC_NOT_FOUND, 
                            "Could not find file " + reqFile );
            //writeToLog(logStr + "\n");
            return;
        }
	if (in == null) {
	  resp.sendError( HttpServletResponse.SC_NOT_FOUND,
				"Could not find file " + reqFile );
	  return;
	}

	t1 = System.currentTimeMillis() - t0; // Mehmet: time for preparing the batch file
	if (traceSuboperationTimes) {
	  //System.out.println("Preparing batch file: " + t1);	// Mehmet: Measure times
	}
	logStr += " " + Long.toString(t1);
	t0 = System.currentTimeMillis();      // Mehmet: Measure times for next suboperation (sending the batch file)

	while( (bytesRead = in.read(buf, 0, buf.length)) > 0 ){
            out.write( buf, 0, bytesRead );
            
            if( Webpp.debugOutput ){
                System.out.write( buf, 0, bytesRead );
            }
        }

	t1 = System.currentTimeMillis() - t0; // Mehmet: time for sending the batch file
	if (traceSuboperationTimes) {		// Mehmet: Measure times
	  //System.out.println("Sending batch file: " + t1);	// Mehmet: Measure times
	}
	logStr += " " + Long.toString(t1);
	t0 = System.currentTimeMillis();      // Mehmet: Measure times for next suboperation (closing the input file and connection)

        try {
            in.close();
        } catch( IOException e ){
          //Assert.assert( false );
	  System.out.println("Error: cannot close the input file: " + e);
        }

	// Measure the overallTime for a GET request
	overallTime = System.currentTimeMillis() - overallTime;
	logStr += " " + Long.toString(overallTime);

	// Write to log file only if client's segment is known
	//  (i.e. request is not from another server or unknown client)
	if (clientSegment != null && !mainReq.endsWith("/")) {
	  writeToLog(logStr + "\n");
	  requestCount++;
	  if ((requestCount % 500) == 0) {
	    System.out.println("Running Garbage Collector");
	    System.gc();
	  }
	}

	// Begin Mehmet
	// If the checksum table is updated, then serialize it to disk
	if (checksumHash.updated())
	  checksumHash.serialize();
	// End Mehmet

    }	// doGet()


   /*
    *=========================================================================
    * Update of a document, no update of replication directory
    * Stx: PUT http://serverIP/servlet/webppserv.Webpp?<urn>&<val2>&..
    * FIXME: Currently, <urn> serves both as a logical URl and as a local
    *        filename. Later we will need to decouple those two.
    *=========================================================================
    */
    protected void doPut( HttpServletRequest req,
                          HttpServletResponse resp )
        throws ServletException, IOException
    {
        ServletInputStream reader = null;
        String storeFile = null;
        FileOutputStream store = null;
        String line = null;
        String qry = null;
        Vector pars;
        String urn = null;
        String urlList = null; 
        byte buffer[] = new byte[4096];
        int len = 0;

	// Mark the beginning of overallTime measurement
        overallTime = System.currentTimeMillis();

        // get qry string
	    qry = req.getQueryString();
	    if( qry == null ){
	        resp.sendError( HttpServletResponse.SC_NOT_FOUND,
	                        "no query string received" );
	        return;
	    }

	// parse qry string
	    pars = Util.split( SubstInputStream.matcher, parmSepPattern, qry );
            for( int i = 0; i < pars.size(); i++ ){
            if( i == 0 ){
                urn = (String) pars.elementAt( i );
            } else if( i == 1 ){
                urlList = (String) pars.elementAt( i );
            } else {
                urlList += " " + (String) pars.elementAt( i );
            }
        }

        // open the stream to read the file to replicate
        reader = req.getInputStream();

        // open the stream to store the file
        storeFile = req.getRealPath( urn );

        if( storeFile == null ){
            throw new ServletException( "could not xlate path " + urn );
        }
        try {
            store  = new FileOutputStream( storeFile );

        } catch( IOException e ){
            resp.sendError( HttpServletResponse.SC_NOT_FOUND,
                            "could not open store file " + storeFile );
	    return;
        }
	if (reader == null || store == null) {
	  resp.sendError( HttpServletResponse.SC_NOT_FOUND,
			"Error while opening store or reader file");
	  return;
	}

        // store the file
        while( (len = reader.read( buffer )) != -1 ){
            store.write( buffer, 0, len );
        }
        store.close();
       
        resp.setStatus( HttpServletResponse.SC_OK,
                        "replicated file " + storeFile );
    }

        
   /*
    *=========================================================================
    * Mehmet: Inter-servlet communication is carried on by using this function.
    * Stx: POST http://serverIP/servlet/webppserv.
    * 'request' gives us the request type, key and value pairs contain the parameters.
    *=========================================================================
    */
    protected void doPost( HttpServletRequest req,
                           HttpServletResponse resp )
        throws ServletException, IOException
    {
	ServletOutputStream out = null;
	String qry = null, reqType = null, parameters = null;
        Vector pars;
	int len = 0;
	byte buffer[] = new byte[4096];
	boolean calledInitInThisRequest = false;

	calledInitInThisRequest = false;
        if( port == -1 ){
            port = req.getServerPort();
        }
        if( rootDir == null ){
            rootDir = req.getRealPath( "/" );
        }
	if( server == null ){
            server = req.getServerName();
            initWebppServer("doPost");          // initialize server
            calledInitInThisRequest = true;
        }

	//System.out.println("ReqType: " + req.getMethod());

	// Mark the beginning of overallTime measurement
        overallTime = System.currentTimeMillis();

	// get qry string
	qry = req.getQueryString();
	if( qry == null ) {
	  resp.sendError( HttpServletResponse.SC_NOT_FOUND,
					"doPost: no query string received" );
	  System.out.println("Error: in doPost no query string received");
	  return;
	}

	System.out.println("Post query: " + qry);
	
	// parse qry string
	pars = Util.split( SubstInputStream.matcher, parmSepPattern, qry );
	for( int i = 0; i < pars.size(); i++ ){
	  if( i == 0 ){
	    reqType = (String) pars.elementAt( i );
	  } else if( i == 1 ){
	    parameters = (String) pars.elementAt( i );
	  } else {
	    parameters += " " + (String) pars.elementAt( i );
	  }
	}

	if (reqType == null) {
	  resp.sendError( HttpServletResponse.SC_NOT_FOUND,
					"doPost: no request type specified");
	  return;
	}

	out = resp.getOutputStream();

	if (reqType.equalsIgnoreCase("ServerList")) {
	  String serverNamesStr = getActiveServerNamesListExcludingNameServer();
	  if (serverNamesStr != null) {
	    //resp.setHeader("ServerList", serverNamesStr);	// Mehmet send the list of server names with port numbers
	    out.print(serverNamesStr);		// Send the list of server names with port numbers
	  }
	}	// ServerList
	else if (reqType.equalsIgnoreCase("GetTestTime")) {
	  String resultStr = Long.toString(System.currentTimeMillis() - testStartTime);
	  if (resultStr != null) {
	    out.print(resultStr);	// Send the time information
	  }
	  System.out.println("Sending test time: " + resultStr);
	}	// GetTestTime
	else if (reqType.equalsIgnoreCase("DownloadReplicationDir")) {
	  DownloadReplicationDir();
	}	// DownloadReplicationDir
	else if (reqType.equalsIgnoreCase("GetReplicationDir")) {
	  SubmitReplicationDir(out);
	  String clientAddr = req.getRemoteAddr();
          String clientName = req.getRemoteHost();
          if (clientAddr != null) {
	    if (clientName != null) {
	      clientName = clientName.toLowerCase();
	      serverStatus.put(clientName, "Alive");
	    }
	    else clientName = "null";
	    writeToLogWithTime("GetReplicationDir " + clientName);
	  }
	}	//GetReplicationDir
	else if (reqType.equalsIgnoreCase("TestPhase")) {
	  String tmp = parameters;	// the remainder of parameters is only the new test phase number
	  testPhase = Integer.valueOf(tmp).intValue();
	  if (testPhase == 0 && !calledInitInThisRequest)
	    initWebppServer("TestPhase");	// initialize server
	  System.out.println("Test Phase: " + tmp);
          writeToLog(Long.toString(System.currentTimeMillis() - testStartTime) + " Server " + server + " TestPhase " + tmp + "\n");
	  //if (testPhase > 2 && serverID != 0) destroy();
	  if (serverID == 0) {
	    broadcastMessage("TestPhase", tmp);	// broadcast to other servers
	    broadcastToClients("TestPhase", tmp);	// broadcast to clients
	    aliveServers = createAliveServersString();
            broadcastMessage("AliveServers", aliveServers, false);
	  }
	}	// TestPhase
	else if (reqType.equalsIgnoreCase("Locations")) {
	  String resourceName = parameters;	// the remainder of parameters is only the resource name
	  String result = ReplicationDir.getUrls(resourceName);
	  out.print(result);	// Mehmet: send the replication directory record
	}	// Locations
	else if (reqType.equalsIgnoreCase("HostServer")) {
	  String clientAddr = req.getRemoteAddr();
	  String clientName = req.getRemoteHost();
	  if (clientName == null) clientName = "null";
	  if (clientAddr != null) {
	    HostServer.put(clientAddr, server);		// put the client into your table
	    System.out.println("Received msg: " + clientAddr + "(" + clientName + ") says I am its host");
            writeToLog(Long.toString(System.currentTimeMillis() - testStartTime) + " Server " + server + " HostServer " + clientAddr + " " + clientName + "\n");
	    broadcastMessage("HostServerBroadcast", clientAddr + " " + server);	// Tell other servers you're this client's host
	    //System.out.println("Broadcasted msg: I am host for " + clientAddr);
	    out.print("Success");
	  }
	}	// HostServer
	else if (reqType.equalsIgnoreCase("HostServerBroadcast")) {
	  String msg = req.getHeader("param");
	  if (msg != null) {
	    StringTokenizer st = new StringTokenizer(msg, " ");
	    if (st.hasMoreElements()) {
	      String clientAddr = (String) st.nextElement();
	      String serverName = null;
	      if (st.hasMoreElements()) {
		serverName = (String) st.nextElement();
		HostServer.put(clientAddr, serverName);		// insert to table
	      
		//System.out.println("Received msg: Host of " + clientAddr + " is " + serverName);
		out.print("Success");
		writeToLogWithTime("HostServerBroadcastIn " + clientAddr + " " + serverName);
	      }
	      else {
		out.print("Failed");
	      }
	    }
	    else {
	      out.print("Failed");
	    }
	  }
	  else {
	    out.print("Failed");
	  }
	}	// HostServerBroadcast
	else if (reqType.equalsIgnoreCase("SendAliveServers")) {
	  // A request for AliveServers info is received
	  // If this is the NAME_SERVER, thenbroadcast alive servers info
	  if (NAME_SERVER.equals(server + ":" + port) ||
			NAME_SERVER_IP.equals(server + ":" + port))
	    aliveServers = createAliveServersString();
	    broadcastMessage("AliveServers", aliveServers, false);
	}
	else if (reqType.equalsIgnoreCase("AliveServers")) {
	  String msg = req.getHeader("param");
	  if (msg != null) {
	    System.out.println("Alive servers received: " + msg);
	    writeToLogWithTime("AliveServers" + " " + msg);
	    StringTokenizer st = new StringTokenizer(msg, " ");
	    while (st.hasMoreElements()) {
	      String serverName = (String) st.nextElement();
	      serverStatus.put(serverName, "Alive");
	      String sPort = (String) st.nextElement();
	      serverPorts.put(serverName, sPort);
              String sID = (String) st.nextElement();
              serverIDs.put(serverName, sID);
	      if (serverName.equals(server)) {
		serverID = Integer.valueOf(sID).intValue();
		port = Integer.valueOf(sPort).intValue();
		//System.out.println("Server-" + serverID + " at " + server + ":" + port + " is up");
	      }
	    }
	    if (segmentationMethod == LATENCY && !latenciesStarted && !latenciesDone)
		doLatencies(this);
	  }
	}	// AliveServers
	else if (reqType.equalsIgnoreCase("Destroy")) {
	  destroyServer();
	}	// Destroy
	else if (reqType.equalsIgnoreCase("ReplicationRequest")) {
	  String msg = req.getHeader("param");
	  if (msg != null) {
	    String requestingServer = null;
            // Mehmet: replaced the next line with the following prgp since
            //  req.getRemoteHost() returns 'begonia.ece.nwu.edu' for the
            //  Northwestern server, instead of 'begonia.ece.northwestern.edu'
            //requestingServer = req.getRemoteHost();

            // Determine the requesting server from the msg string
            int pos, pos2;
            pos = msg.indexOf("//");
            if (pos == -1) {
              System.out.println("Error: cannot parse requestingServer in ReplicationRequest request msg");
            }
            else {
              pos2 = msg.indexOf('/', pos + 2);
              if (pos2 == -1) {
                System.out.println("Error: cannot parse the end of requestingServer in ReplicationRequest request msg");
              }
              else {
                requestingServer = msg.substring(pos+2, pos2);
              }
	    }

	    if (requestingServer != null) requestingServer = requestingServer.toLowerCase();
	    System.out.println("Replication request: " + requestingServer + " wants to replicate " + msg + " on this server");
	    pos = msg.indexOf(' ');
	    if (pos == -1 || requestingServer == null) {
	      System.out.println("Error: cannot parse filename and url in replication request msg");
	      out.print("Failed");
	    }
	    else {	// msg successfully parsed
	      String fname = msg.substring(0, pos);
	      String url = msg.substring(pos + 1);
	      if (serverContainsResource(server, fname)) {
		//System.out.println("Warning: I already have a replica of " + fname);
		out.print("Failed");
		writeToLog(Long.toString(System.currentTimeMillis() - testStartTime) + " Server " + server + " ReplicaRejected " + fname + " " + requestingServer + "\n");
	      }
	      else {	// it is okay to do the replication
		downloadResource(requestingServer, fname, url);
		out.print("Success");
		writeToLog(Long.toString(System.currentTimeMillis() - testStartTime) + " Server " + server + " ReplicaAccepted " + fname + " " + requestingServer + "\n");
	      }
	    }
	  }
	}	// ReplicationRequest
	else if (reqType.equalsIgnoreCase("AddToReplDir")) {
	  String msg = req.getHeader("param");
	  if (msg != null) {
	    String requestingServer = null;
	    // Mehmet: replaced the next line with the following prgp since
	    // 	req.getRemoteHost() returns 'begonia.ece.nwu.edu' for the
	    //	Northwestern server, instead of 'begonia.ece.northwestern.edu'
	    //requestingServer = req.getRemoteHost();

	    // Determine the requesting server from the msg string
	    int pos, pos2;
	    pos = msg.indexOf("//");
	    if (pos == -1) {
              System.out.println("Error: cannot parse requestingServer in AddToReplDir request msg");
            } 
	    else {
	      pos2 = msg.indexOf('/', pos + 2);
	      if (pos2 == -1) {
		System.out.println("Error: cannot parse the end of requestingServer in AddToReplDir request msg");
	      }
	      else {
		requestingServer = msg.substring(pos+2, pos2);
	      }
	    }

	    if (requestingServer != null) requestingServer = requestingServer.toLowerCase();
	    System.out.println("AddToReplDir request received: " + requestingServer + " adds " +
			msg + " to replication directory");
	    pos = msg.indexOf(' ');
	    if (pos == -1 || requestingServer == null) {
	      System.out.println("Error: cannot parse filename and url in AddToReplDir request msg");
	      out.print("Failed");
	    }
	    else {	// msg successfully parsed
	      String fname = msg.substring(0, pos);
	      String url = msg.substring(pos + 1);
	      repl.appendUrl(fname, url);
	      out.print("Success");
	      writeToLogWithTime("AddToReplDir " + fname + " " + requestingServer);
	    }
	  }
	}	// AddToReplDir

	out.close();
    }	// doPost()
    
    
   /*
    *=========================================================================
    * Destructor
    *=========================================================================
    */    
    public void destroy()
    {
        cache.close();
        repl.close();
	checksumHash.close();
	writeToLogWithTime("End");
	try {
	  logfile.close();
	} catch (IOException e) {
	  System.out.println("Error: cannot close log file");
	}
	//super.destroy();
    }

   /*
    *=========================================================================
    * Submit the whole replication directory to the requesting server
    *=========================================================================
   */
    public void SubmitReplicationDir(ServletOutputStream out)
    {
	File fd = null;
        BufferedReader in = null;
        String line = null;
	String serFileName = null;
	File replDirFd;
	String replDirName;

	replDirName = webppRootDir +
                      System.getProperty( "file.separator" ) +
                      "replication";
	replDirFd = new File( replDirName );

	serFileName = replDirName + System.getProperty( "file.separator" ) +
                      "replicationDir.ser";
        fd = new File( serFileName );
	if( fd.exists() ){
            /* deserialize an existing replication directory */
	  if( debugReplDir ){
                System.out.println( "***: deserializing " + serFileName );
            }

            try {
                in = new BufferedReader( new FileReader( fd ) );
            } catch( Exception e ){
                Assert.assert( false, e.getMessage() );
            }

            try {
                while( (line = in.readLine()) != null ){
                    if( debugReplDir ){
                        System.out.println( line );
                    }

                    if( line.indexOf( ' ' ) == -1 ){
                        continue;
                    }

		    out.print(line + "\n");
                }
            } catch( Exception e ){
                Assert.assert( false, e.getMessage() );
            }
          try {
              in.close();
          } catch( Exception e ){
              Assert.assert( false, e.getMessage() );
          }
	}
	else {
	  System.out.println("Error: cannot open Replication Directory in SubmitReplicationDir");
	  return;
	}
    }  // SubmitReplicationDir()

   /*
    *=========================================================================
    * Retrieve the whole replication directory from the NAME_SERVER 
    *=========================================================================
   */
    public void DownloadReplicationDir()
    {
        String tmp = null, lurl, purl, line;
	int pos;

        String url = NAME_SERVER;
        try {
          tmp = postURL(url, "GetReplicationDir", null);
        } catch (Exception e) {
          tmp = null;
          System.out.println("Error: cannot send GetReplicationDir request to NAME_SERVER: " + e);
        }
        // If name server cannot be contacted, also try its IP address
        if (tmp == null) {
          url = NAME_SERVER_IP;
          try {
            tmp = postURL(url, "GetReplicationDir", null);
          } catch (Exception e) {
            tmp = null;
            System.out.println("Error: cannot send GetReplicationDir request to NAME_SERVER: " + e);
          }
        }

	if (tmp == null) return;

	StringTokenizer st = new StringTokenizer(tmp, "\n");
        for ( ; st.hasMoreElements(); ) {
	  line = (String) st.nextElement();
	  pos = line.indexOf( ' ' );
          if( pos == -1 ){
            continue;
          }
	  lurl = line.substring( 0, pos );
	  purl = line.substring( pos+1, line.length() );
	  if( lurl.length() == 0 || purl.length() == 0 ){
	    continue;
	  }
	  repl.put( lurl, purl );
	}
	repl.serialize();
	writeToLogWithTime("GotReplicationDir");
    }	// DownloadReplicationDir()

   /*
    *=========================================================================
    * Get the current test time from the NAME_SERVER
    *=========================================================================
   */
    public long getTestTimeFromNameServer()
    {
        long result = 0, tmplong;
        String tmp = null;

        String url = NAME_SERVER;
        try {
          tmp = postURL(url, "GetTestTime", null);
        } catch (Exception e) {
          tmp = null;
          System.out.println("Error: cannot send GetTestTime request to NAME_SERVER: " + e);
        }
        // If name server cannot be contacted, also try its IP address
        if (tmp == null) {
          url = NAME_SERVER_IP;
          try {
            tmp = postURL(url, "GetTestTime", null);
          } catch (Exception e) {
            tmp = null;
            System.out.println("Error: cannot send GetTestTime request to NAME_SERVER: " + e);
          }
	}

        if (tmp == null) return 0;
        tmplong = (new Long(tmp)).longValue();
	System.out.println("Received test time: " + tmp);
        result = System.currentTimeMillis() - tmplong;
        return result;
    }	// getTestTimeFromNameServer()
    
   /*
    *=========================================================================
    * Deduce mime type from file suffix (the Apache way)
    *=========================================================================
    */        
    protected String getMimeType( String fileName )
    {
        String mimeType;
        
        mimeType = getServletConfig().getServletContext().
                   getMimeType( fileName );
                   
        if( mimeType == null ){
            return "text/plain";
        }
        
        return mimeType;
        
        /*
        dotIx = fileName.lastIndexOf( '.' );
        
        if( dotIx == -1 ){
            return "text/plain";
        }
        
        s = fileName.substring( dotIx+1 );
        */
        /********************************************************************
         FIXME: at the time the list is not complete, more types should be
                put in for a working prototype.
         ********************************************************************/
        /*
        if( s.equals("html") || s.equals("htm") ){
            return "text/html";
        }
        
        if( s.equals("txt") ){
            return "text/plain";
        }
        
        if( s.equals("mpeg") || s.equals("mpg") || s.equals("mpe") ){
            return "video/mpeg";
        }
        
        if( s.equals("gif") ){
            return "image/gif";
        }
        
        if( s.equals("jpeg") || s.equals("jpg") || s.equals("jpe") ){
            return "image/jpeg";
        }
        
        if( s.equals("png") ){
            return "image/png";
        }
        
        if( s.equals("tiff") || s.equals("tif") ){
            return "image/tiff";
        }
        
        if( s.equals("ai") || s.equals("eps") || s.equals("ps") ){
            return "application/postscript";
        }  
        
        if( s.equals("ppt") ){
            return "application/powerpoint";
        }
        
        if( s.equals("pdf") ){
            return "application/pdf";
        }
        
        if( s.equals("doc") ){
            return "application/msword";
        }
        
        if( s.equals("zip") ){
            return "application/zip";
        }
        
        return "text/plain";
        */
    }

   /*
    *=========================================================================
    * getChecksumValue: Calculates the checksum for the file whose name is given as argument.
    * Mehmet: this fcn is called while parsing the parent HTML document for embedded resources.
    *=========================================================================
   */

    static String getChecksumValue(Checksum checksum, String fname) {
	String result = (String) checksumHash.get(fname);	// is this checksum already calculated during this request's service
	if (result != null) {
	  //System.out.println("Checksum found in hash table for " + fname);
	  return result;	// return the checksum that was already calculated and stored in hash table
	}
        //System.out.println("Calculating checksum for " + fname);
        long t0 = System.currentTimeMillis();		// Mehmet: Measure times
        try {
            BufferedInputStream is = new BufferedInputStream(
                new FileInputStream(fname));
            byte[] bytes = new byte[1024];
            int len = 0;
            
            // Read the file and compute the checksum.
            while ((len = is.read(bytes)) >= 0) {
                checksum.update(bytes, 0, len);
            }
            is.close();
        } catch (IOException e) {
            //e.printStackTrace();
            netChecksumTime += System.currentTimeMillis() - t0;	// Mehmet: Measure times
	    return null;
        }
        long chksm = checksum.getValue();
	result = new String();
	result = result.valueOf(chksm);
	netChecksumTime += System.currentTimeMillis() - t0;	// Mehmet: Measure times
	checksumHash.put(fname, result);
	return result;
    }


   /*
    *=========================================================================
    * serverIsAlive: returns true if that server is alive, false otherwise
    *=========================================================================
    */
    static boolean serverIsAlive(String serverName)
    {
      if (serverName.equals(server)) return true;
      String serverStat = (String) serverStatus.get(serverName);
      if (serverStat != null && serverStat.equalsIgnoreCase("Alive")) return true;
      else return false;
    }

   /*
    *=========================================================================
    * clientIsAlive: returns true if that client is alive, false otherwise
    *=========================================================================
    */
    static boolean clientIsAlive(String serverName)
    {
      if (serverName.equals(server)) return true;
      String serverStat = (String) clientStatus.get(serverName);
      if (serverStat != null && serverStat.equalsIgnoreCase("Alive")) return true;
      else return false;
    }

   /*
    *=========================================================================
    * createAliveServersString: collects information about active Web++
    *  servers into one string object. Information includes server name,
    *  port, and ID number.
    *=========================================================================
    */
    static String createAliveServersString()
    {
	Enumeration enum = serverPorts.keys();
	String serverName = null, serverNamesStr = null;
	while( enum.hasMoreElements() ) {
	  serverName = (String) enum.nextElement();
	  if (!serverIsAlive(serverName)) continue;     // skip if that server is dead
	  serverName += " " + ((String) serverPorts.get(serverName)) +
			" " + ((String) serverIDs.get(serverName));
          if (serverNamesStr == null)
            serverNamesStr = new String(serverName);
          else
            serverNamesStr += " " + serverName;
        }
        return serverNamesStr;
    }

   /*
    *=========================================================================
    * getActiveServerNamesListExcludingNameServer: collects active
    *  Web++ server names into one string object, but excludes
    *  the name of NAME_SERVER
    *=========================================================================
    */
    static String getActiveServerNamesListExcludingNameServer()
    {
      Enumeration enum = serverPorts.keys();
      String serverName = null, serverNamesStr = null;
      while( enum.hasMoreElements() ) {
          serverName = (String) enum.nextElement();
          if (!serverIsAlive(serverName)) continue;     // skip if that server is dead
          serverName += ":" + ((String) serverPorts.get(serverName));
	  // The following if statement is the only difference of
	  //  this function from getActiveServerNamesList() function.
	  if (NAME_SERVER.equals(serverName) ||
                        NAME_SERVER_IP.equals(serverName))
	    continue;
          if (serverNamesStr == null)
            serverNamesStr = new String(serverName);
          else
            serverNamesStr += " " + serverName;
        }
        return serverNamesStr;
    }	// getActiveServerNamesListExcludingNameServer()

   /*
    *=========================================================================
    * getActiveServerNamesList: collects active Web++ server names into one string object
    *=========================================================================
    */
    static String getActiveServerNamesList()
    {
      Enumeration enum = serverPorts.keys();
      String serverName = null, serverNamesStr = null;
      while( enum.hasMoreElements() ) {
	  serverName = (String) enum.nextElement();
	  if (!serverIsAlive(serverName)) continue;	// skip if that server is dead
	  serverName += ":" + ((String) serverPorts.get(serverName));
	  if (serverNamesStr == null)
	    serverNamesStr = new String(serverName);
	  else
	    serverNamesStr += " " + serverName;
	}
	return serverNamesStr;
    }	// getActiveServerNamesList()

   /*
    *=========================================================================
    * getServerNamesList: collects all Web++ server names into one string object
    *=========================================================================
    */
    static String getServerNamesList()
    {
      Enumeration enum = serverPorts.keys();
      String serverName = null, serverNamesStr = null;
      while( enum.hasMoreElements() ) {
	serverName = (String) enum.nextElement();
	serverName += ":" + ((String) serverPorts.get(serverName));
	if (serverNamesStr == null)
	  serverNamesStr = new String(serverName);
	else
	  serverNamesStr += " " + serverName;
	}
	return serverNamesStr;
    }	// getServerNamesList

   /*
    *=========================================================================
    * readOptionsFile: reads a few initialization options from a file
    *   The information is inserted into some Webpp static variables.
    *=========================================================================
    */

    static void readOptionsFile()
    {
	BufferedReader in = null;
        File fd = null;
        String line = null, str1, str2;
        int pos1;
        String dirName =  Webpp.webppRootDir +
                      System.getProperty( "file.separator" ) +
                      "replication";
        File dirFd = new File( dirName );

        if( !dirFd.exists() )
          Assert.assert( true,"Cannot find directory: " + dirName );

        String serFileName = dirName + System.getProperty( "file.separator" ) +
                                                        "init.ser";
        fd = new File( serFileName );

        if( fd.exists() ){
          try {
            in = new BufferedReader( new FileReader( fd ) );
          } catch( Exception e ){
            Assert.assert( false, e.getMessage() );
          }
	  try {
            while( (line = in.readLine()) != null ){
              pos1 = line.indexOf( ' ' );
              if( pos1 == -1 ) continue;
              str1 = line.substring( 0, pos1 );
              str2 = line.substring( pos1+1 );
	      if (str1.equalsIgnoreCase("MinReqCountForDecision")) {
		MinReqCountForDecision = Long.valueOf(str2).longValue();
	      }
	      else if (str1.equalsIgnoreCase("MinLoadPercentageForDecision")) {
		MinLoadPercentageForDecision = Double.valueOf(str2).doubleValue();
	      }
	      else if (str1.equalsIgnoreCase("MinReqPercentageForHotDoc")) {
		MinReqPercentageForHotDoc = Double.valueOf(str2).doubleValue();
	      }
	      else if (str1.equalsIgnoreCase("LimitToEnforceClearCount")) {
		LimitToEnforceClearCount = Long.valueOf(str2).longValue();
	      }
            }
          } catch( Exception e ){
            Assert.assert( false, e.getMessage() );
          }
        }
        else {
          Assert.assert( true,"Cannot find file: " + serFileName );
        }

    }	// readOptionsFile()

   /*
    *=========================================================================
    * readServerList: reads list of servers and info about each server.
    *   The information is inserted into serverIDs and serverPorts hash
    *   tables.
    *=========================================================================
    */

    static void readServerList()
    {
	BufferedReader in = null;
	File fd = null;
	String line = null, str1, str2;
	int pos1, pos2;
	String dirName =  Webpp.webppRootDir +
                      System.getProperty( "file.separator" ) +
                      "replication";
	File dirFd = new File( dirName );

	if( !dirFd.exists() )
	  Assert.assert( true,"Cannot find directory: " + dirName );
      
	String serFileName = dirName + System.getProperty( "file.separator" ) +
							"servers.ser";
	fd = new File( serFileName );
			
	if( fd.exists() ){
	  try {
            in = new BufferedReader( new FileReader( fd ) );
          } catch( Exception e ){
            Assert.assert( false, e.getMessage() );
          }
          try {
	    int nextID = 0;
            while( (line = in.readLine()) != null ){
	      if (line.startsWith("#")) continue;	// skip commented lines
              pos1 = line.indexOf( ' ' );
              if( pos1 == -1 ) continue;
              str1 = line.substring( 0, pos1 );
	      str2 = String.valueOf(nextID);
              serverIDs.put(str1, str2);
              pos2 = line.length();
              str2 = line.substring( pos1+1, pos2 );
              serverPorts.put(str1, str2);
	      nextID++;
            }
          } catch( Exception e ){
	    Assert.assert( false, e.getMessage() );
	  }
	}
	else {
	  Assert.assert( true,"Cannot find file: " + serFileName );
        }
    }

       /*
    *=========================================================================
    * readClientList: reads list of clients and info about each client.
    *   The information is inserted into clientIDs and clientPorts hash
    *   tables.
    *=========================================================================
    */

    static void readClientList()
    {
        BufferedReader in = null;
        File fd = null;
        String line = null, str1, str2;
        int pos1, pos2;
        String dirName =  Webpp.webppRootDir +
                      System.getProperty( "file.separator" ) +
                      "replication";
        File dirFd = new File( dirName );

        if( !dirFd.exists() )
          Assert.assert( true,"Cannot find directory: " + dirName );
	String serFileName = dirName + System.getProperty( "file.separator" ) +
                                                        "clients.ser";
        fd = new File( serFileName );

        if( fd.exists() ){
          try {
            in = new BufferedReader( new FileReader( fd ) );
          } catch( Exception e ){
            Assert.assert( false, e.getMessage() );
          }
          try {
            int nextID = 0;
            while( (line = in.readLine()) != null ){
	      if (line.startsWith("#")) continue;	// skip commented lines
              pos1 = line.indexOf( ' ' );
              if( pos1 == -1 ) continue;
              str1 = line.substring( 0, pos1 );
              str2 = String.valueOf(nextID);
              clientIDs.put(str1, str2);
              pos2 = line.length();
              str2 = line.substring( pos1+1, pos2 );
              clientPorts.put(str1, str2);
              nextID++;
	    }
	  } catch( Exception e ){
            Assert.assert( false, e.getMessage() );
          }
        }
        else {
          Assert.assert( true,"Cannot find file: " + serFileName );
        }
    }	// readClientList()




    
    /*
    *=========================================================================
    * initServers: Sends dummy requests to other servers so that they start
    *   constructing the server clusters. Then, it calls the function at the
    *   local server to start the same operation.
    *=========================================================================
    */

    static void initServers(String caller)
    {
	Enumeration enum;
	String serverName = null;
	GetURL myGetURL = new GetURL();
	long bytesRead;
	
	serverStatus.put(server, "Alive");	// put my status as Alive

	enum = serverPorts.keys();
	while( enum.hasMoreElements() ) {
	  bytesRead = 0;
	  serverName = (String) enum.nextElement();
	  if (serverName == null) {
	    System.out.println("Error: server name is null in initServers");
	    return;
	  }
	  //System.out.println("Initializing server: " + serverName);
	  if (serverName.equals(server)) continue;	// don't call yourself
	  try {
	    bytesRead = myGetURL.getURL("http://" + serverName + ":" +
	    		serverPorts.get(serverName) + 
	    		"/servlet/webppserv.Webpp/indexWebpp.html");
	  } catch (Exception e) {
	    System.out.println("Error while calling server " + serverName + " " + e);
	    serverStatus.put(serverName, "Dead");
	    bytesRead = 0;
	  }
	  if (bytesRead > 0) {
	    System.out.println(serverName + " is contacted successfully");
	    serverStatus.put(serverName, "Alive");
	  }
	}
	//System.out.println("Done checking servers. Now will broadcast live servers");
	aliveServers = createAliveServersString();
	// If caller is TestPhase then do not bcast AliveServers,
	//  because it will be broadcasted later
	if (caller == null || !caller.equals("TestPhase"))
	  broadcastMessage("AliveServers", aliveServers, false);
	serversInitialized = true;
    }

    static void initClients(Webpp app)
    {
	Runnable r = new ClientInitiator(app);
        Thread t = new Thread(r);
        t.start();
        return;
    }

    static void initClients_old()
    {
        Enumeration enum;
        String clientName = null;
        GetURL myGetURL = new GetURL();
        long bytesRead;

        enum = clientPorts.keys();
        while( enum.hasMoreElements() ) {
          bytesRead = 0;
          clientName = (String) enum.nextElement();
          if (clientName == null) {
            System.out.println("Error: client name is null in initClients");
            return;
          }
	  //System.out.println("Initializing client: " + clientName);
          if (clientName.equals(server)) continue;      // don't call yourself
          try {
            bytesRead = myGetURL.getURL("http://" + clientName + ":" +
                        clientPorts.get(clientName) +
                        "/servlet/webppclient.Webpp/indexWebpp.html");
          } catch (Exception e) {
            System.out.println("Error while initiating client " + clientName);
            clientStatus.put(clientName, "Dead");
            bytesRead = 0;
          }
          if (bytesRead > 0) {
            System.out.println(clientName + " is contacted successfully");
            clientStatus.put(clientName, "Alive");
          }
        }
	//System.out.println("Done initiating clients.");
    }	// initClients()

    /*
    *=========================================================================
    * doTraceroutes: Sends traceroute requests to other servers in order to
    *   construct the tree structure view of network topology and create
    *   segments of servers on that tree structure.
    *=========================================================================
    */

    static void doTraceroutes(Webpp parentServlet)
    {
	traceroutesStarted = true;
	// If segments are not built, build or read from disk
	if (assigned == null)
	  assigned = SerHashtable.factory("segments.ser", parentServlet);
	if (!assigned.isNew()) return;	// if it was deserialized, cancel segment building operations
	Runnable r = new SegmentBuilder();
	Thread t = new Thread(r);
	t.start();
	return;
    }


    /*
    *=========================================================================
    * doLatencies: Assigns each server to a separate segment by itself.
    *   Then, it sends dummy requests to other servers in order to
    *   measure their distances to the current server.
    *=========================================================================
    */

    static void doLatencies(Webpp parentServlet)
    {
	//System.out.println("Latency measurement started");
	latenciesStarted = true;
	// If segments are not built, build or read from disk
	if (assigned == null)
	  assigned = SerHashtable.factory("segments.ser", parentServlet);
	if (!assigned.isNew()) return;	// if it was deserialized, cancel segment building operations

	// Segments were not found on disk, so build new segments. Assign each server
	//	to a segment by its own (to simplify, assign each segment to a segment
	//	numbered the same as the serverID number of that server).
	latenciesStarted = true;
	Enumeration enum = serverIDs.keys();
	String serverName = null;
	Integer segment = null, segSize;
	int seg = -1;
        while( enum.hasMoreElements() ) {
	  serverName = (String) enum.nextElement();
	  if (serverName == null || !serverIsAlive(serverName)) continue;
	  segment = new Integer((String) serverIDs.get(serverName));
	  if (segment == null) continue;
	  if (seg < (segment.intValue() + 1)) seg = segment.intValue() + 1;
	  assigned.put(serverName, segment);
	  segSize = new Integer(1);
          segmentSizes.put(segment, segSize);
	  System.out.println("Server " + serverName + " assigned to segment " + segment);
	}
	noOfSegments = seg;	// set Webpp global variable
	if (respTimes != null) {
	  respTimes.setNoOfSegments(seg);
	}
	assigned.serialize();

	// Start the latency distance measurement process
	//Runnable r = new LatencyMeasurer();
	//Thread t = new Thread(r);
	//t.start();
	return;
    }

    /*
    *=========================================================================
    * extractFilenameFromURL: extracts filename portion from the given URL
    *=========================================================================
    */

    static String extractFilenameFromURL(String url)
    {
	String result = null;
	int pos;
	pos = url.lastIndexOf("/");
	if (pos == -1) {
	  System.out.println("Error: cannot parse requestingServer in extractFilenameFromURL");
	}
	else {
	  result = url.substring(pos+1);
	}
	return result;
    }

    /*
    *=========================================================================
    * extractServerNamePortFromURL: extracts server name and port portion from
    *  the given URL
    *=========================================================================
    */

    static String extractServerNamePortFromURL(String url)
    {
	String result = null;
	int pos, pos2;
	pos = url.indexOf("//");
	if (pos == -1) {
	  System.out.println("Error: cannot parse requestingServer in extractServerNamePortFromURL");
        }
	else {
	  pos2 = url.indexOf('/', pos + 2);
	  if (pos2 == -1) {
	    System.out.println("Error: cannot parse the end of requestingServer in extractServerNamePortFromURL");
	  }
	  else {
	    result = url.substring(pos+2, pos2);
	  }
	}
	if (result != null) result = result.toLowerCase();
	return result;
    }

   /*
    *=========================================================================
    * broadcastMessage: Sends a broadcast message to all alive servers.
    *  The parameters of the broadcast msg are included both in the URL
    *  and the headers of the request.
    */

    static void broadcastMessage(String msgType, String param)
    {
	broadcastMessage(msgType, param, true);
    }

    /*
    *=========================================================================
    * broadcastMessage: Sends a broadcast message to all alive servers.
    *  If useURL is set to true, then the parameters of the broadcast
    *  msg are included both in the URL and the headers of the request.
    *  If useURL is set to false, then the parameters of the broadcast
    *  msg are only included in the header of the request.
    *=========================================================================
    */

    static void broadcastMessage(String msgType, String param, boolean useURL)
    {
	Runnable r = new MessageBroadcaster(msgType, param, "Servers", useURL);
	Thread t = new Thread(r);
	t.start();
	if (!msgType.equals("AddToReplDir"))
	  writeToLogWithTime("Bcast" + msgType + " " + param);
	else {
	  String tmp = extractServerNamePortFromURL(param);
	  String tmp2 = extractFilenameFromURL(param);
	  if (tmp != null && tmp2 != null)
	    writeToLogWithTime("Bcast" + msgType + " " + tmp2 + " " + tmp);
	}
	return;
    }

    /*
    *=========================================================================
    * broadcastToClients: Sends a broadcast message to all alive clients 
    *=========================================================================
    */

    static void broadcastToClients(String msgType, String param)
    {
        Runnable r = new MessageBroadcaster(msgType, param, "Clients");
        Thread t = new Thread(r);
        t.start();
	writeToLogWithTime("BCastClients" + msgType + " " + param);
        return;
    }

   /*
    *=========================================================================
    * isThisAServer: returns true if the given name corresponds to a server
    *=========================================================================
    */
    static boolean thisIsAServer(String addr, String name)
    {
	boolean result = false;
	if (addr != null) result = serverIsAlive(addr);
	if (name!= null && !result) {
	  result = serverIsAlive(name);
	  if (!result) result = serverIsAlive(name.toLowerCase());
	}
	return result;
    }

   /*
    *=========================================================================
    * getSegmentofClient: finds out which segment the given client belongs to.
    *=========================================================================
    */
    static Integer getSegmentofClient(String client, String clientName)
    {
      if (client == null || assigned == null || HostServer == null || HostServer.isEmpty()) return null;
      boolean clientIsAServer = thisIsAServer(client, clientName);
      String hostServer = (String) HostServer.get(client);
      if (hostServer == null) {
	if (!clientIsAServer)
	  System.out.println("Warning: do not know the host server of " + client + "(" + clientName + ")");
	return null;
      }
      if (assigned == null || !assigned.containsKey(hostServer)) {
	if (!clientIsAServer)
	  System.out.println("Warning: do not know the segment of server " + hostServer);
	return null;
      }
      Integer seg = (Integer) assigned.get(hostServer);
      return seg;
    }

   /*
    *=========================================================================
    * insertRespTime: inserts the response time info into respTimes hash table
    *=========================================================================
    */
    static void insertRespTime(Integer clientSeg, String prevRespTimeStr)
    {
      if (clientSeg == null) return;
      int seg = clientSeg.intValue();
      int pos = prevRespTimeStr.indexOf('|');
      if (pos == -1) {
        System.out.println("Error: cannot separate fields of prevRespTimeStr");
        return;
      }
      String fileName = prevRespTimeStr.substring(0, pos);
      String tmp = prevRespTimeStr.substring(pos + 1);
      if (tmp == null) {
        System.out.println("Error: cannot get response time out of prevRespTimeStr");
        return;
      }
      long rtime = (new Long(tmp)).longValue();
      //System.out.println("Inserting resp time: " + fileName + " seg: " + seg + " time: " + rtime);
      respTimes.put(fileName, seg, rtime);
      checkForReplication(fileName);
      if (totalRequestCount >= LimitToEnforceClearCount)
	respTimes.clear();	// Enforce a clear of all request counts
    }

    /*
    *=========================================================================
    * checkForReplication: performs a check to decide whether a new replica of the
    *	given resource is needed or not
    *=========================================================================
    */
    static void checkForReplication(String fname)
    {
	//System.out.println("Check for Replication: " + fname);
	if (assigned == null) return;	// segments not built yet
	if (!respTimes.containsKey(fname)) return;
	if (!isHotResource(fname)) return;	// don't replicate if not a hot resource
	double maxPercentage = respTimes.getMostActiveSegmentPercentage(fname);
	if (maxPercentage < MinLoadPercentageForDecision)
	  return;	// don't replicate if there are no overloaded segments

	// Now decided to replicate. Get the segment number.
	Integer segment = respTimes.getMostActiveSegment(fname);
	//System.out.println("I decided to replicate " + fname + " in segment " + segment);

	// Find an alive target server in the chosen segment
	String serverName = null;
	Vector candidateServers = new Vector();
	Enumeration enum = serverPorts.keys();
	while( enum.hasMoreElements() ) {
	  serverName = (String) enum.nextElement();
	  Integer seg = (Integer) assigned.get(serverName);
	  if (seg == null || !seg.equals(segment)) continue;	// skip if not in chosen segment
	  if (!serverIsAlive(serverName)) continue;	// skip if that server is dead
	  if (serverContainsResource(serverName, fname)) continue;	// skip if that server already has a replica
	  //serverName += ":" + ((String) serverPorts.get(serverName));
	  candidateServers.addElement(serverName);	// add server into candidates list
	}
	int chosenIndex;
	if (candidateServers.size() == 0) {
	  //System.out.println("Warning: cannot find a server for a new replica in segment " + segment);
	  return;
	}
	else if (candidateServers.size() == 1)
	  chosenIndex = 0;
	else {
	  // More than one candidate servers, choose one randomly
	  Random rnd = new Random(System.currentTimeMillis());
	  chosenIndex = rnd.nextInt() % candidateServers.size();
	}

	String chosenServer = (String) candidateServers.elementAt(chosenIndex);
	if (chosenServer == null) {
	  //System.out.println("Error: chosenServer is null");
	  return;
	}

	//System.out.println("No of Candidate Servers = " + candidateServers.size());
	//System.out.println("ChosenIndex = " + chosenIndex);
	System.out.println("Chosen server for replication: " + chosenServer);
	writeToLog(Long.toString(System.currentTimeMillis() - testStartTime) + " Server " + server + " ReplicationDecision " + fname + " " + segment + " " + chosenServer + "\n");

	// Perform replication, update replication directory, broadcast the directory.
	// First, get the target server's port number.
	String targetPort = (String) serverPorts.get(chosenServer);
	if (targetPort == null) {
	  System.out.println("Error: cannot figure out the port number of " + chosenServer);
	  return;
	}

	// Get the URL of local replica so that target server can request it from current server.
	String localURL = repl.getLocalURL(fname);
	if (localURL == null) {
	  System.out.println("Error: cannot figure out the local URL of " + fname);
	  return;
	}

	// Send the replication request to the target server, and leave the rest to it.
	try {
	  postURL(chosenServer + ":" + targetPort, "ReplicationRequest", fname + " " + localURL);
	} catch (Exception e) {
	  System.out.println("Error while sending ReplicationRequest to " + chosenServer);
	  return;
	}

	// Clear counts to restart with fresh request counts
	respTimes.clear();
    }

    /*
    *=========================================================================
    * isHotResource performs a check to decide whether a resource is hot
    *	enough to be considered for replication, returns true if so
    *=========================================================================
    */
    static boolean isHotResource(String fname)
    {
	boolean answer = true;
	long count = respTimes.getRequestCount(fname);

	if (count < MinReqCountForDecision)
	  return false;		// don't replicate without enough information
	
	if (count < totalRequestCount * MinReqPercentageForHotDoc)
	  return false;

	System.out.println("Hot file detected: " + fname);
	return answer;
    }

    /*
    *=========================================================================
    * serverContainsResource: checks the replication directory to see if the
    *	given server already contains a replica of the given resource
    *=========================================================================
    */
    static boolean serverContainsResource(String serverName, String fname)
    {
	String replList = ReplicationDir.getUrls(fname);
	if (replList.indexOf(serverName) != -1) {
	  //System.out.println("Warning: Server " + serverName + " already has a replica of " + fname);
	  return true;
	}
	return false;
    }

    /*
    *=========================================================================
    * postURL: send a POST request to the given server
    *=========================================================================
    */
    static String postURL(String serverName, String msgType, String param) throws Exception
    {
	byte b = 0;
	int done = 0, bytesRead;
	String urlName = "http://" + serverName + "/servlet/webppserv.Webpp/?" + msgType;
	URL url = new URL(urlName);
	HttpURLConnection conn = (HttpURLConnection) url.openConnection();
	String resultStr = null;
	StringBuffer mybuf = new StringBuffer(256);
	InputStream tmpin = null;
    
	conn.setRequestMethod("POST");
	bytesRead = 0;

	// Send your Post message
	if (param != null)
	  conn.setRequestProperty("param", param);

	// Get the response
	tmpin = conn.getInputStream();
	if (tmpin == null) return null;
	DataInputStream in = new DataInputStream(tmpin);
	if (in == null) return null;
	try { b = in.readByte(); } catch (EOFException e) { done = 1; }
	while (done == 0) {
	  bytesRead++;
	  mybuf.append((char)b);
	  try { b = in.readByte(); } catch (EOFException e) { done = 1; }
	}
	in.close();
	if (mybuf.length() < 1) return null;
	resultStr = new String(mybuf);
	return resultStr;
    }

    /*
    *=========================================================================
    * downloadResource: download a copy of the given resource from the given server
    *	at the given URL
    *=========================================================================
    */
    static void downloadResource(String remoteServer, String fname, String url)
    {
	Runnable r = new ResourceDownloader(remoteServer, fname, url);
	Thread t = new Thread(r);
	t.start();
	return;
    }

    /*
    *=========================================================================
    * getNoOfSegmentsFromTable: finds out the largest segment number in the
    *	'assigned' table. The no of segments is (largest seg no + 1). So, this
    *	function returns (largest seg no + 1) as the no of segments.
    *=========================================================================
    */
    static int getNoOfSegmentsFromTable()
    {
	if (assigned == null) return 0;
	int maxSeg = -1, seg;
	Enumeration enum = assigned.elements();
	while( enum.hasMoreElements() ) {
	  Integer sno = (Integer) enum.nextElement();
	  seg = sno.intValue();
	  if (seg > maxSeg) maxSeg = seg;
	}
	if (maxSeg == -1) {
	  System.out.println("Error: cannot determine the noOfSegments from the table");
	  return 0;
	}
	return maxSeg + 1;
    }

}	// class ends here

final class ClientInitiator extends Thread
{
    private Webpp applet_;

    // constructor ============================================================
    ClientInitiator( Webpp app )
    {
	// remember reference to parent applet
        applet_ = app;
    }

    public void run()
    {
	Enumeration enum;
        String clientName = null;
        GetURL myGetURL = new GetURL();
        long bytesRead;

	try { setPriority(MIN_PRIORITY); } catch (Exception ee) {}
	yield();	// give a chance to the caller thread to continue
	// Wait 30 seconds before initiating the clients
        try {
          sleep(30000);
        } catch (InterruptedException ie) {}

        enum = applet_.clientPorts.keys();
        while( enum.hasMoreElements() ) {
          bytesRead = 0;
          clientName = (String) enum.nextElement();
          if (clientName == null) {
            System.out.println("Error: client name is null in initClients");
            return;
          }
          //System.out.println("Initializing client: " + clientName);
          if (clientName.equals(applet_.server)) continue;      // don't call yourself
          try {
            bytesRead = myGetURL.getURL("http://" + clientName + ":" +
                        applet_.clientPorts.get(clientName) +
                        "/servlet/webppclient.Webpp/indexWebpp.html");
          } catch (Exception e) {
            System.out.println("Error while initiating client " + clientName);
            applet_.clientStatus.put(clientName, "Dead");
            bytesRead = 0;
          }
          if (bytesRead > 0) {
            System.out.println(clientName + " is contacted successfully");
            applet_.clientStatus.put(clientName, "Alive");
          }
	  yield();	// give parent thread a chance to run
        }	// while
        //System.out.println("Done initiating clients.");
    }	// run()
}
